/* Copyright (C) 2007 Andre Noll <maan@tuebingen.mpg.de>, see file COPYING. */

/**
 * \file afs.c Paraslash's audio file selector (afs).
 *
 * The functions of this file execute either in afs or in command handler
 * process context where the afs and the command handler processes are siblings
 * of each other, i.e. direct descendants of the server process. Inter-process
 * communication between command handlers, server and afs is performed through
 * Unix domain sockets.
 */

#include <netinet/in.h>
#include <sys/socket.h>
#include <signal.h>
#include <fnmatch.h>
#include <osl.h>
#include <lopsub.h>
#include <arpa/inet.h>
#include <sys/un.h>
#include <netdb.h>
#include <lopsub.h>

#include "server.lsg.h"
#include "server_cmd.lsg.h"
#include "para.h"
#include "error.h"
#include "crypt.h"
#include "string.h"
#include "afh.h"
#include "afs.h"
#include "net.h"
#include "server.h"
#include "daemon.h"
#include "ipc.h"
#include "list.h"
#include "sched.h"
#include "fd.h"
#include "signal.h"
#include "sideband.h"
#include "command.h"

/**
 * The array of tables of the audio file selector.
 *
 * We organize them in an array to be able to loop over all tables.
 */
static const struct afs_table {
	/** The name is no table operation, so define it here. */
	const char * const name;
	/** The only way to invoke the ops is via this pointer. */
	const struct afs_table_operations *ops;
} afs_tables[] = {
	{.name = "audio_files", .ops = &aft_ops},
	{.name = "attributes", .ops = &attr_ops},
	{.name = "scores", .ops = &score_ops},
	{.name = "moods", .ops = &moods_ops},
	{.name = "lyrics", .ops = &lyrics_ops},
	{.name = "images", .ops = &images_ops},
	{.name = "playlists", .ops = &playlists_ops},
};
/** Used to loop over the afs tables. */
#define NUM_AFS_TABLES ARRAY_SIZE(afs_tables)

struct command_task {
	/** The file descriptor for the local socket. */
	int fd;
	/** The associated task structure. */
	struct task *task;
};

extern int mmd_mutex;
extern struct misc_meta_data *mmd;

static int server_socket;
static struct command_task command_task_struct;
static struct signal_task *signal_task;

static enum play_mode current_play_mode, previous_play_mode;
static char *current_mop, *previous_mop; /* NULL means dummy mood */

extern uint32_t afs_socket_cookie;

/**
 * Passed from command handlers to afs.
 *
 * Command handlers cannot change the afs database directly because they run in
 * a separate process. The callback query structure circumvents this
 * restriction as follows. To instruct the afs process to execute a particular
 * function, the command hander writes an instance of this structure to a
 * shared memory area, along with the arguments to the callback function. The
 * identifier of the shared memory area is transferred to the afs process via
 * the command socket.
 *
 * The afs process reads the shared memory id from the command socket, attaches
 * the corresponding area, and calls the callback function whose address is
 * stored in the area.
 *
 * The command output, if any, is transferred back to the command handler in
 * the same way: The afs process writes the output to a second shared memory
 * area together with a fixed size metadata header whose format corresponds to
 * the \ref callback_result structure. The identifier of this area is sent back
 * to the command handler which attaches the area and forwards the output to
 * the remote client.
 *
 * \sa \ref struct callback_result.
 */
struct callback_query {
	/** The function to be called. */
	afs_callback *cb;
	/** The number of bytes of the query */
	size_t query_size;
};

/**
 * Structure embedded in the result of a callback.
 *
 * If the callback produced a result, an instance of that structure is embedded
 * into the shared memory area holding the result, mainly to let the command
 * handler know the size of the result.
 *
 * \sa \ref struct callback_query.
 */
struct callback_result {
	/** The number of bytes of the result. */
	size_t result_size;
	/** The band designator (loglevel for the result). */
	uint8_t band;
};

static int dispatch_result(int result_shmid, callback_result_handler *handler,
		void *private_result_data)
{
	struct osl_object result;
	void *result_shm;
	/* must attach r/w as result.data might get encrypted in-place. */
	int ret2, ret = shm_attach(result_shmid, ATTACH_RW, &result_shm);
	struct callback_result *cr = result_shm;

	if (ret < 0) {
		PARA_ERROR_LOG("attach failed: %s\n", para_strerror(-ret));
		return ret;
	}
	result.size = cr->result_size;
	result.data = result_shm + sizeof(*cr);
	assert(handler);
	ret = handler(&result, cr->band, private_result_data);
	ret2 = shm_detach(result_shm);
	if (ret2 < 0) {
		PARA_ERROR_LOG("detach failed: %s\n", para_strerror(-ret2));
		if (ret >= 0)
			ret = ret2;
	}
	return ret;
}

/**
 * Ask the afs process to call a given function.
 *
 * \param f The function to be called.
 * \param query Pointer to arbitrary data for the callback.
 * \param result_handler Called for each shm area sent by the callback.
 * \param private_result_data Passed verbatim to \a result_handler.
 *
 * This function creates a socket for communication with the afs process and a
 * shared memory area (sma) to which the buffer pointed to by \a query is
 * copied.  It then notifies the afs process that the callback function \a f
 * should be executed by sending the shared memory identifier (shmid) to the
 * socket.
 *
 * If the callback produces a result, it sends any number of shared memory
 * identifiers back via the socket. For each such identifier received, \a
 * result_handler is called. The contents of the sma identified by the received
 * shmid are passed to that function as an osl object. The private_result_data
 * pointer is passed as the second argument to \a result_handler.
 *
 * \return Number of shared memory areas dispatched on success, negative on
 * errors.
 */
int send_callback_request(afs_callback *f, struct osl_object *query,
		callback_result_handler *result_handler,
		void *private_result_data)
{
	struct callback_query *cq;
	int ret, fd = -1, query_shmid, result_shmid;
	void *query_shm;
	char buf[sizeof(afs_socket_cookie) + sizeof(int)];
	size_t query_shm_size = sizeof(*cq);
	int dispatch_error = 0, num_dispatched = 0;

	if (query)
		query_shm_size += query->size;
	ret = shm_new(query_shm_size);
	if (ret < 0)
		return ret;
	query_shmid = ret;
	ret = shm_attach(query_shmid, ATTACH_RW, &query_shm);
	if (ret < 0)
		goto out;
	cq = query_shm;
	cq->cb = f;
	cq->query_size = query_shm_size - sizeof(*cq);

	if (query)
		memcpy(query_shm + sizeof(*cq), query->data, query->size);
	ret = shm_detach(query_shm);
	if (ret < 0)
		goto out;

	*(uint32_t *)buf = afs_socket_cookie;
	*(int *)(buf + sizeof(afs_socket_cookie)) = query_shmid;

	ret = connect_local_socket(OPT_STRING_VAL(AFS_SOCKET));
	if (ret < 0)
		goto out;
	fd = ret;
	ret = write_all(fd, buf, sizeof(buf));
	if (ret < 0)
		goto out;
	/*
	 * Read all shmids from afs.
	 *
	 * Even if the dispatcher returns an error we _must_ continue to read
	 * shmids from fd so that we can destroy all shared memory areas that
	 * have been created for us by the afs process.
	 */
	for (;;) {
		ret = recv_bin_buffer(fd, buf, sizeof(int));
		if (ret <= 0)
			goto out;
		assert(ret == sizeof(int));
		ret = *(int *) buf;
		assert(ret > 0);
		result_shmid = ret;
		ret = dispatch_result(result_shmid, result_handler,
			private_result_data);
		if (ret < 0 && dispatch_error >= 0)
			dispatch_error = ret;
		ret = shm_destroy(result_shmid);
		if (ret < 0)
			PARA_CRIT_LOG("destroy result failed: %s\n",
				para_strerror(-ret));
		num_dispatched++;
	}
out:
	if (shm_destroy(query_shmid) < 0)
		PARA_CRIT_LOG("shm destroy error\n");
	if (fd >= 0)
		close(fd);
	if (dispatch_error < 0)
		return dispatch_error;
	if (ret < 0)
		return ret;
	return num_dispatched;
}

/**
 * Wrapper for send_callback_request() which passes a lopsub parse result.
 *
 * \param f The callback function.
 * \param cmd Needed for (de-)serialization.
 * \param lpr Must match cmd.
 * \param private_result_data Passed to send_callback_request().
 *
 * This function serializes the parse result given by the lpr pointer into a
 * buffer. The buffer is sent as the query to the afs process with the callback
 * mechanism.
 *
 * \return The return value of the underlying call to send_callback_request().
 */
int send_lls_callback_request(afs_callback *f,
		const struct lls_command * const cmd,
		struct lls_parse_result *lpr, void *private_result_data)
{
	struct osl_object query;
	char *buf = NULL;
	int ret = lls_serialize_parse_result(lpr, cmd, &buf, &query.size);

	assert(ret >= 0);
	query.data = buf;
	ret = send_callback_request(f, &query, afs_cb_result_handler,
		private_result_data);
	free(buf);
	return ret;
}

static int action_if_pattern_matches(struct osl_row *row, void *data)
{
	struct pattern_match_data *pmd = data;
	struct osl_object name_obj;
	const char *p, *name;
	int i, ret;

	ret = osl(osl_get_object(pmd->table, row, pmd->match_col_num,
		&name_obj));
	if (ret < 0)
		return ret;
	name = (char *)name_obj.data;
	if ((!name || !*name) && (pmd->pm_flags & PM_SKIP_EMPTY_NAME))
		return 1;
	if (lls_num_inputs(pmd->lpr) == 0) {
		if (pmd->pm_flags & PM_NO_PATTERN_MATCHES_EVERYTHING) {
			pmd->num_matches++;
			return pmd->action(pmd->table, row, name, pmd->data);
		}
	}
	i = pmd->input_skip;
	for (;;) {
		if (i >= lls_num_inputs(pmd->lpr))
			break;
		p = lls_input(i, pmd->lpr);
		ret = fnmatch(p, name, pmd->fnmatch_flags);
		if (ret != FNM_NOMATCH) {
			if (ret != 0)
				return -E_FNMATCH;
			ret = pmd->action(pmd->table, row, name, pmd->data);
			if (ret >= 0)
				pmd->num_matches++;
			return ret;

		}
		i++;
	}
	return 1;
}

/**
 * Execute the given function for each matching row.
 *
 * \param pmd Describes what to match and how.
 *
 * \return Standard.
 */
int for_each_matching_row(struct pattern_match_data *pmd)
{
	if (pmd->pm_flags & PM_REVERSE_LOOP)
		return osl(osl_rbtree_loop_reverse(pmd->table, pmd->loop_col_num, pmd,
			action_if_pattern_matches));
	return osl(osl_rbtree_loop(pmd->table, pmd->loop_col_num, pmd,
			action_if_pattern_matches));
}

/**
 * Compare two osl objects of string type.
 *
 * \param obj1 Pointer to the first object.
 * \param obj2 Pointer to the second object.
 *
 * In any case, only \p MIN(obj1->size, obj2->size) characters of each string
 * are taken into account.
 *
 * \return It returns an integer less than, equal to, or greater than zero if
 * \a obj1 is found, respectively, to be less than, to match, or be greater than
 * obj2.
 *
 * \sa strcmp(3), strncmp(3).
 */
int string_compare(const struct osl_object *obj1, const struct osl_object *obj2)
{
	const char *str1 = obj1->data;
	const char *str2 = obj2->data;
	return strncmp(str1, str2, PARA_MIN(obj1->size, obj2->size));
}

static int pass_afd(int fd, char *buf, size_t size)
{
	struct msghdr msg = {.msg_iov = NULL};
	struct cmsghdr *cmsg;
	char control[255] __a_aligned(8);
	int ret;
	struct iovec iov;

	iov.iov_base = buf;
	iov.iov_len  = size;

	msg.msg_iov = &iov;
	msg.msg_iovlen = 1;

	msg.msg_control = control;
	msg.msg_controllen = sizeof(control);

	cmsg = CMSG_FIRSTHDR(&msg);
	cmsg->cmsg_level = SOL_SOCKET;
	cmsg->cmsg_type = SCM_RIGHTS;
	cmsg->cmsg_len = CMSG_LEN(sizeof(int));
	*(int *)CMSG_DATA(cmsg) = fd;

	/* Sum of the length of all control messages in the buffer */
	msg.msg_controllen = cmsg->cmsg_len;
	PARA_DEBUG_LOG("passing %zu bytes and fd %d\n", size, fd);
	ret = sendmsg(server_socket, &msg, 0);
	if (ret < 0) {
		ret = -ERRNO_TO_PARA_ERROR(errno);
		return ret;
	}
	return 1;
}

/*
 * Send audio file information to the server process.
 *
 * This first calls \ref open_and_update_audio_file() to obtain a file
 * descriptor to the upcoming audio file and an id of a shared memory area
 * containing information for the virtual streaming system. The file descriptor
 * and the id are transferred to the server process through the Unix domain
 * socket that was passed to \ref afs_init().
 */
static int open_next_audio_file(void)
{
	int ret, shmid, fd;
	char buf[8];

	ret = open_and_update_audio_file(&fd);
	if (ret < 0) {
		if (ret != -OSL_ERRNO_TO_PARA_ERROR(E_OSL_RB_KEY_NOT_FOUND))
			PARA_ERROR_LOG("%s\n", para_strerror(-ret));
		goto no_admissible_files;
	}
	shmid = ret;
	if (!write_ok(server_socket)) {
		ret = -E_AFS_SOCKET;
		goto destroy;
	}
	*(uint32_t *)buf = NEXT_AUDIO_FILE;
	*(uint32_t *)(buf + 4) = (uint32_t)shmid;
	ret = pass_afd(fd, buf, 8);
	close(fd);
	if (ret >= 0)
		return ret;
destroy:
	shm_destroy(shmid);
	return ret;
no_admissible_files:
	*(uint32_t *)buf = NO_ADMISSIBLE_FILES;
	*(uint32_t *)(buf + 4) = (uint32_t)0;
	return write_all(server_socket, buf, 8);
}

static int activate_mood_or_playlist(const char *arg, struct para_buffer *pb)
{
	enum play_mode mode;
	int ret;
	char *msg;

	if (!arg) { /* load dummy mood */
		ret = mood_load(NULL, NULL, &msg);
		mode = PLAY_MODE_MOOD;
	} else if (!strcmp(arg, "-")) { /* load previous mop */
		char *mop = previous_mop? previous_mop + 2 : NULL;
		if (previous_play_mode == PLAY_MODE_MOOD)
			ret = mood_load(mop, NULL, &msg);
		else
			ret = playlist_load(mop, NULL, &msg);
		mode = previous_play_mode;
	} else if (!strncmp(arg, "p/", 2)) {
		ret = playlist_load(arg + 2, NULL, &msg);
		mode = PLAY_MODE_PLAYLIST;
	} else if (!strncmp(arg, "m/", 2)) {
		ret = mood_load(arg + 2, NULL, &msg);
		mode = PLAY_MODE_MOOD;
	} else {
		ret = -ERRNO_TO_PARA_ERROR(EINVAL);
		msg = make_message("%s: parse error\n", arg);
	}
	if (pb)
		para_printf(pb, "%s", msg);
	free(msg);
	if (ret < 0)
		return ret;
	/*
	 * We get called with arg == current_mop from the signal dispatcher
	 * after SIGHUP and from the error path of the select command to
	 * re-select the current mood or playlist. Don't update the four global
	 * mop variables in these cases.
	 */
	if (arg != current_mop) {
		previous_play_mode = current_play_mode;
		current_play_mode = mode;
		if (arg && !strcmp(arg, "-")) {
			char *tmp = current_mop;
			current_mop = previous_mop;
			previous_mop = tmp;
		} else {
			free(previous_mop);
			previous_mop = current_mop;
			current_mop = arg? para_strdup(arg) : NULL;
		}
	}
	/* Notify the server about the mood/playlist change. */
	mutex_lock(mmd_mutex);
	snprintf(mmd->afs_mode_string, sizeof(mmd->afs_mode_string) - 1,
		"%s", current_mop? current_mop : "dummy");
	mmd->afs_mode_string[sizeof(mmd->afs_mode_string) - 1] = '\0';
	mmd->events++;
	mutex_unlock(mmd_mutex);
	return 1;
}

/**
 * Result handler for sending data to the para_client process.
 *
 * \param result The data to be sent.
 * \param band The band designator.
 * \param private Pointer to the command context.
 *
 * \return The return value of the underlying call to \ref command.c::send_sb.
 *
 * \sa \ref callback_result_handler, \ref command.c::send_sb.
 */
int afs_cb_result_handler(struct osl_object *result, uint8_t band,
		void *private)
{
	struct command_context *cc = private;

	assert(cc);
	switch (band) {
	case SBD_OUTPUT:
	case SBD_DEBUG_LOG:
	case SBD_INFO_LOG:
	case SBD_NOTICE_LOG:
	case SBD_WARNING_LOG:
	case SBD_ERROR_LOG:
	case SBD_CRIT_LOG:
	case SBD_EMERG_LOG:
		assert(result->size > 0);
		return send_sb(&cc->scc, result->data, result->size, band, true);
	case SBD_AFS_CB_FAILURE:
		return *(int *)(result->data);
	default:
		return -E_BAD_BAND;
	}
}

static void flush_and_free_pb(struct para_buffer *pb)
{
	int ret;
	struct afs_max_size_handler_data *amshd = pb->private_data;

	if (pb->buf && pb->size > 0) {
		ret = pass_buffer_as_shm(amshd->fd, amshd->band, pb->buf,
			pb->offset);
		if (ret < 0)
			PARA_ERROR_LOG("%s\n", para_strerror(-ret));
	}
	free(pb->buf);
}

static void init_admissible_files(const char *arg)
{
	int ret = activate_mood_or_playlist(arg, NULL);
	if (ret < 0) {
		PARA_WARNING_LOG("could not activate %s: %s\n", arg?
			arg : "dummy", para_strerror(-ret));
		if (arg)
			activate_mood_or_playlist(NULL, NULL);
	}
}

static int setup_command_socket_or_die(void)
{
	int ret, socket_fd;
	const char *socket_name = OPT_STRING_VAL(AFS_SOCKET);

	unlink(socket_name);
	ret = create_local_socket(socket_name);
	if (ret < 0) {
		PARA_EMERG_LOG("%s: %s\n", para_strerror(-ret), socket_name);
		exit(EXIT_FAILURE);
	}
	socket_fd = ret;
	PARA_INFO_LOG("listening on socket %s (fd %d)\n", socket_name,
		socket_fd);
	return socket_fd;
}

static char *database_dir;

static void close_afs_tables(void)
{
	int i;
	PARA_NOTICE_LOG("closing afs tables\n");
	for (i = 0; i < NUM_AFS_TABLES; i++)
		afs_tables[i].ops->close();
	free(database_dir);
	database_dir = NULL;
}

static void get_database_dir(void)
{
	if (!database_dir) {
		if (OPT_GIVEN(AFS_DATABASE_DIR))
			database_dir = para_strdup(OPT_STRING_VAL(AFS_DATABASE_DIR));
		else {
			char *home = para_homedir();
			database_dir = make_message(
				"%s/.paraslash/afs_database-0.7", home);
			free(home);
		}
	}
	PARA_INFO_LOG("afs_database dir %s\n", database_dir);
}

static int open_afs_tables(void)
{
	int i, ret;

	get_database_dir();
	PARA_NOTICE_LOG("opening %zu osl tables in %s\n", NUM_AFS_TABLES,
		database_dir);
	for (i = 0; i < NUM_AFS_TABLES; i++) {
		ret = afs_tables[i].ops->open(database_dir);
		if (ret >= 0)
			continue;
		PARA_ERROR_LOG("could not open %s\n", afs_tables[i].name);
		break;
	}
	if (ret >= 0)
		return ret;
	while (i)
		afs_tables[--i].ops->close();
	return ret;
}

static int afs_signal_post_monitor(struct sched *s, __a_unused void *context)
{
	int signum, ret;

	if (getppid() == 1) {
		PARA_EMERG_LOG("para_server died\n");
		goto shutdown;
	}
	signum = para_next_signal();
	if (signum == 0)
		return 0;
	if (signum == SIGHUP) {
		close_afs_tables();
		parse_config_or_die(1);
		ret = open_afs_tables();
		if (ret < 0)
			return ret;
		init_admissible_files(current_mop);
		return 0;
	}
	PARA_EMERG_LOG("terminating on signal %d\n", signum);
shutdown:
	task_notify_all(s, E_AFS_SIGNAL);
	return -E_AFS_SIGNAL;
}

static void register_signal_task(struct sched *s)
{
	para_sigaction(SIGPIPE, SIG_IGN);
	signal_task = signal_init_or_die();
	para_install_sighandler(SIGINT);
	para_install_sighandler(SIGTERM);
	para_install_sighandler(SIGHUP);

	signal_task->task = task_register(&(struct task_info) {
		.name = "signal",
		.pre_monitor = signal_pre_monitor,
		.post_monitor = afs_signal_post_monitor,
		.context = signal_task,

	}, s);
}

static struct list_head afs_client_list;

/** Describes one connected afs client. */
struct afs_client {
	/** Position in the afs client list. */
	struct list_head node;
	/** The socket file descriptor for this client. */
	int fd;
	/** The time the client connected. */
	struct timeval connect_time;
};

static void command_pre_monitor(struct sched *s, void *context)
{
	struct command_task *ct = context;
	struct afs_client *client;

	sched_monitor_readfd(server_socket, s);
	sched_monitor_readfd(ct->fd, s);
	list_for_each_entry(client, &afs_client_list, node)
		sched_monitor_readfd(client->fd, s);
}

/**
 * Send data as shared memory to a file descriptor.
 *
 * \param fd File descriptor to send the shmid to.
 * \param band The band designator for this data.
 * \param buf The buffer holding the data to be sent.
 * \param size The size of \a buf.
 *
 * This function creates a shared memory area large enough to hold
 * the content given by \a buf and \a size and sends the identifier
 * of this area to the file descriptor \a fd.
 *
 * It is called by the AFS max_size handler as well as directly by the AFS
 * command callbacks to send command output to the command handlers.
 *
 * \return Zero if \a buf is \p NULL or \a size is zero. Negative on errors,
 * and positive on success.
 */
int pass_buffer_as_shm(int fd, uint8_t band, const char *buf, size_t size)
{
	int ret, shmid;
	void *shm;
	struct callback_result *cr;

	if (size == 0)
		assert(band != SBD_OUTPUT);
	ret = shm_new(size + sizeof(*cr));
	if (ret < 0)
		return ret;
	shmid = ret;
	ret = shm_attach(shmid, ATTACH_RW, &shm);
	if (ret < 0)
		goto err;
	cr = shm;
	cr->result_size = size;
	cr->band = band;
	if (size > 0)
		memcpy(shm + sizeof(*cr), buf, size);
	ret = shm_detach(shm);
	if (ret < 0)
		goto err;
	ret = write_all(fd, (char *)&shmid, sizeof(int));
	if (ret >= 0)
		return ret;
err:
	if (shm_destroy(shmid) < 0)
		PARA_ERROR_LOG("destroy result failed\n");
	return ret;
}

/**
 * Format and send an error message to the command handler.
 *
 * To pass an error message from the callback of an afs command to the client,
 * this function should be called. It formats the message into a buffer which
 * is passed as a shared memory area to the command handler from where it
 * propagates to the client.
 *
 * The message will be tagged with the ERROR_LOG sideband designator so that
 * the client writes it to its stderr stream rather than to stdout as with
 * aca->pbout. In analogy to the default Unix semantics of stderr, the message
 * is sent without buffering.
 *
 * If sending the error message fails, an error is logged on the server side,
 * but no other action is taken.
 *
 * \param aca Used to obtain the fd to send the shmid to.
 * \param fmt Usual format string.
 */
__printf_2_3 void afs_error(const struct afs_callback_arg *aca,
		const char *fmt,...)
{
	va_list argp;
	char *msg;
	unsigned n;
	int ret;

	va_start(argp, fmt);
	n = xvasprintf(&msg, fmt, argp);
	va_end(argp);
	ret = pass_buffer_as_shm(aca->fd, SBD_ERROR_LOG, msg, n + 1);
	if (ret < 0)
		PARA_ERROR_LOG("Could not send %s: %s\n", msg,
			para_strerror(-ret));
	free(msg);
}

static int call_callback(int fd, int query_shmid)
{
	void *query_shm;
	struct callback_query *cq;
	int ret, ret2;
	struct afs_callback_arg aca = {.fd = fd};

	ret = shm_attach(query_shmid, ATTACH_RW, &query_shm);
	if (ret < 0)
		return ret;
	cq = query_shm;
	aca.query.data = (char *)query_shm + sizeof(*cq);
	aca.query.size = cq->query_size;
	aca.pbout.max_size = shm_get_shmmax();
	aca.pbout.max_size_handler = afs_max_size_handler;
	aca.pbout.private_data = &(struct afs_max_size_handler_data) {
		.fd = fd,
		.band = SBD_OUTPUT
	};
	ret = cq->cb(&aca);
	ret2 = shm_detach(query_shm);
	if (ret2 < 0) {
		if (ret < 0) /* ignore (but log) detach error */
			PARA_ERROR_LOG("could not detach sma: %s\n",
				para_strerror(-ret2));
		else
			ret = ret2;
	}
	flush_and_free_pb(&aca.pbout);
	if (ret < 0) {
		ret2 = pass_buffer_as_shm(fd, SBD_AFS_CB_FAILURE,
			(const char *)&ret, sizeof(ret));
		if (ret2 < 0)
			PARA_ERROR_LOG("could not pass cb failure packet: %s\n",
				para_strerror(-ret));
	}
	return ret;
}

static int execute_server_command(void)
{
	char buf[8];
	size_t n;
	int ret = read_nonblock(server_socket, buf, sizeof(buf) - 1, &n);

	if (ret < 0 || n == 0)
		return ret;
	buf[n] = '\0';
	if (strcmp(buf, "new"))
		return -ERRNO_TO_PARA_ERROR(EINVAL);
	return open_next_audio_file();
}

/* returns 0 if no data available, 1 else */
static int execute_afs_command(int fd)
{
	uint32_t cookie;
	int query_shmid;
	char buf[sizeof(cookie) + sizeof(query_shmid)];
	size_t n;
	int ret = read_nonblock(fd, buf, sizeof(buf), &n);

	if (ret < 0)
		goto err;
	if (n == 0)
		return 0;
	if (n != sizeof(buf)) {
		PARA_NOTICE_LOG("short read (%d bytes, expected %lu)\n",
			ret, (long unsigned) sizeof(buf));
		return 1;
	}
	cookie = *(uint32_t *)buf;
	if (cookie != afs_socket_cookie) {
		PARA_NOTICE_LOG("received invalid cookie (got %u, expected %u)\n",
			(unsigned)cookie, (unsigned)afs_socket_cookie);
		return 1;
	}
	query_shmid = *(int *)(buf + sizeof(cookie));
	if (query_shmid < 0) {
		PARA_WARNING_LOG("received invalid query shmid %d)\n",
			query_shmid);
		return 1;
	}
	ret = call_callback(fd, query_shmid);
	if (ret >= 0)
		return 1;
err:
	PARA_NOTICE_LOG("%s\n", para_strerror(-ret));
	return 1;
}

/** Shutdown connection if query has not arrived until this many seconds. */
#define AFS_CLIENT_TIMEOUT 3

static int command_post_monitor(struct sched *s, void *context)
{
	struct command_task *ct = context;
	struct sockaddr_un unix_addr;
	struct afs_client *client, *tmp;
	int fd, ret;

	ret = task_get_notification(ct->task);
	if (ret < 0)
		return ret;
	ret = execute_server_command();
	if (ret < 0) {
		PARA_EMERG_LOG("%s\n", para_strerror(-ret));
		task_notify_all(s, -ret);
		return ret;
	}
	/* Check the list of connected clients. */
	list_for_each_entry_safe(client, tmp, &afs_client_list, node) {
		ret = execute_afs_command(client->fd);
		if (ret == 0) { /* prevent bogus connection flooding */
			struct timeval diff;
			tv_diff(now, &client->connect_time, &diff);
			if (diff.tv_sec < AFS_CLIENT_TIMEOUT)
				continue;
			PARA_WARNING_LOG("connection timeout\n");
		}
		close(client->fd);
		list_del(&client->node);
		free(client);
	}
	/* Accept connections on the local socket. */
	ret = para_accept(ct->fd, &unix_addr, sizeof(unix_addr), &fd);
	if (ret < 0)
		PARA_NOTICE_LOG("%s\n", para_strerror(-ret));
	if (ret <= 0)
		return 0;
	ret = mark_fd_nonblocking(fd);
	if (ret < 0) {
		PARA_NOTICE_LOG("%s\n", para_strerror(-ret));
		close(fd);
		return 0;
	}
	client = alloc(sizeof(*client));
	client->fd = fd;
	client->connect_time = *now;
	para_list_add(&client->node, &afs_client_list);
	return 0;
}

static void register_command_task(struct sched *s)
{
	struct command_task *ct = &command_task_struct;
	ct->fd = setup_command_socket_or_die();

	ct->task = task_register(&(struct task_info) {
		.name = "afs command",
		.pre_monitor = command_pre_monitor,
		.post_monitor = command_post_monitor,
		.context = ct,
	}, s);
}

static int afs_poll(struct pollfd *fds, nfds_t nfds, int timeout)
{
	mutex_lock(mmd_mutex);
	daemon_set_loglevel(mmd->loglevel);
	mutex_unlock(mmd_mutex);
	return xpoll(fds, nfds, timeout);
}
/**
 * Initialize the audio file selector.
 *
 * \param socket_fd File descriptor for communication with the server process.
 *
 * Open all tables of the afs database, then schedule the signal task and the
 * command task. The former reacts to signals by reloading the database or by
 * terminating after closing all database tables. The command task dispatches
 * commands from the server and from command handlers.
 *
 * The server connects the command task when the virtual streaming system
 * needs to stream the next audio file. The command task opens the highest
 * scoring admissible file and sends streaming information back to the
 * server. The command task also accepts connections from command handlers,
 * modifies or queries the afs database accordingly, and reports its findings
 * back to the command handler.
 */
__noreturn void afs_init(int socket_fd)
{
	static struct sched s;
	int ret;

	register_signal_task(&s);
	init_list_head(&afs_client_list);
	ret = open_afs_tables();
	if (ret < 0)
		goto out;
	server_socket = socket_fd;
	ret = mark_fd_nonblocking(server_socket);
	if (ret < 0)
		goto out_close;
	PARA_INFO_LOG("server_socket: %d\n", server_socket);
	init_admissible_files(OPT_STRING_VAL(AFS_INITIAL_MODE));
	register_command_task(&s);
	s.default_timeout = 1000;
	s.poll_function = afs_poll;
	ret = write(socket_fd, "\0", 1);
	if (ret != 1) {
		if (ret == 0)
			errno = EINVAL;
		ret = -ERRNO_TO_PARA_ERROR(errno);
		goto out_close;
	}
	ret = schedule(&s);
	sched_shutdown(&s);
	mood_unload(NULL);
	playlist_unload(NULL);
out_close:
	close_afs_tables();
out:
	signal_shutdown(signal_task);
	free_status_items();
	free(current_mop);
	free(previous_mop);
	free_lpr();
	if (ret < 0)
		PARA_EMERG_LOG("%s\n", para_strerror(-ret));
	exit(EXIT_FAILURE);
}

static int com_select_callback(struct afs_callback_arg *aca)
{
	const struct lls_command *cmd = SERVER_CMD_CMD_PTR(SELECT);
	const char *arg;
	int ret;
	struct para_buffer *pbout;

	ret = lls_deserialize_parse_result(aca->query.data, cmd, &aca->lpr);
	assert(ret >= 0);
	arg = lls_input(0, aca->lpr);
	pbout = SERVER_CMD_OPT_GIVEN(SELECT, VERBOSE, aca->lpr)?
		&aca->pbout : NULL;
	score_clear();
	if (current_play_mode == PLAY_MODE_MOOD)
		mood_unload(NULL);
	else
		playlist_unload(NULL);
	ret = activate_mood_or_playlist(arg, pbout);
	if (ret >= 0)
		goto free_lpr;
	/* ignore subsequent errors (but log them) */
	if (current_mop && strcmp(current_mop, arg) != 0) {
		int ret2;
		afs_error(aca, "switching back to %s\n", current_mop);
		ret2 = activate_mood_or_playlist(current_mop, pbout);
		if (ret2 >= 0)
			goto free_lpr;
		afs_error(aca, "could not reactivate %s: %s\n", current_mop,
			para_strerror(-ret2));
	}
	activate_mood_or_playlist(NULL, pbout);
free_lpr:
	lls_free_parse_result(aca->lpr, cmd);
	return ret;
}

static int com_select(struct command_context *cc, struct lls_parse_result *lpr)
{
	const struct lls_command *cmd = SERVER_CMD_CMD_PTR(SELECT);
	char *errctx;
	int ret = lls(lls_check_arg_count(lpr, 1, 1, &errctx));

	if (ret < 0) {
		send_errctx(cc, errctx);
		return ret;
	}
	ret = send_lls_callback_request(com_select_callback, cmd, lpr, cc);
	return ret == osl(-E_OSL_RB_KEY_NOT_FOUND)? -E_BAD_MOP : ret;
}
EXPORT_SERVER_CMD_HANDLER(select);

static int com_init_callback(struct afs_callback_arg *aca)
{
	uint32_t table_mask = *(uint32_t *)aca->query.data;
	int i, ret;

	close_afs_tables();
	get_database_dir();
	for (i = 0; i < NUM_AFS_TABLES; i++) {
		const struct afs_table *t = afs_tables + i;

		if (!(table_mask & (1 << i)))
			continue;
		if (!t->ops->create)
			continue;
		ret = t->ops->create(database_dir);
		if (ret < 0) {
			afs_error(aca, "cannot create table %s\n", t->name);
			goto out;
		}
		para_printf(&aca->pbout, "successfully created %s table\n",
			t->name);
	}
	ret = open_afs_tables();
	if (ret < 0)
		afs_error(aca, "cannot open afs tables: %s\n",
			para_strerror(-ret));
out:
	return ret;
}

static int com_init(struct command_context *cc, struct lls_parse_result *lpr)
{
	int i, j, ret;
	uint32_t table_mask = (1 << (NUM_AFS_TABLES + 1)) - 1;
	struct osl_object query = {.data = &table_mask,
		.size = sizeof(table_mask)};
	unsigned num_inputs = lls_num_inputs(lpr);

	get_database_dir();
	ret = para_mkdir(database_dir);
	if (ret < 0)
		return ret;
	if (num_inputs > 0) {
		table_mask = 0;
		for (i = 0; i < num_inputs; i++) {
			for (j = 0; j < NUM_AFS_TABLES; j++) {
				const struct afs_table *t = afs_tables + j;

				if (strcmp(lls_input(i, lpr), t->name))
					continue;
				table_mask |= (1 << j);
				break;
			}
			if (j == NUM_AFS_TABLES)
				return -E_BAD_TABLE_NAME;
		}
	}
	return send_callback_request(com_init_callback, &query,
		afs_cb_result_handler, cc);
}
EXPORT_SERVER_CMD_HANDLER(init);

static int com_check(struct command_context *cc, struct lls_parse_result *lpr)
{
	const struct lls_opt_result *r_a = SERVER_CMD_OPT_RESULT(CHECK, AFT, lpr);
	const struct lls_opt_result *r_A = SERVER_CMD_OPT_RESULT(CHECK, ATTRIBUTE, lpr);
	const struct lls_opt_result *r_m = SERVER_CMD_OPT_RESULT(CHECK, MOOD, lpr);
	const struct lls_opt_result *r_p = SERVER_CMD_OPT_RESULT(CHECK, PLAYLIST, lpr);
	bool noopt = !lls_opt_given(r_a) && !lls_opt_given(r_A)
		&& !lls_opt_given(r_m) && !lls_opt_given(r_p);
	int ret;

	if (noopt || lls_opt_given(r_a)) {
		ret = send_callback_request(aft_check_callback, NULL,
			afs_cb_result_handler, cc);
		if (ret < 0)
			return ret;
	}
	if (noopt || lls_opt_given(r_A)) {
		ret = send_callback_request(attribute_check_callback, NULL,
			afs_cb_result_handler, cc);
		if (ret < 0)
			return ret;
	}
	if (noopt || lls_opt_given(r_p)) {
		ret = send_callback_request(playlist_check_callback,
			NULL, afs_cb_result_handler, cc);
		if (ret < 0)
			return ret;
	}
	if (noopt || lls_opt_given(r_m)) {
		ret = send_callback_request(mood_check_callback, NULL,
			afs_cb_result_handler, cc);
		if (ret < 0)
			return ret;
	}
	return 1;
}
EXPORT_SERVER_CMD_HANDLER(check);

/**
 * The afs event dispatcher.
 *
 * \param event Type of the event.
 * \param pb May be \p NULL.
 * \param data Type depends on \a event.
 *
 * This function calls each table event handler, passing \a pb and \a data
 * verbatim. It's up to the handlers to interpret the \a data pointer. If a
 * handler returns negative, the loop is aborted.
 *
 * \return The (negative) error code of the first handler that failed, or non-negative
 * if all handlers succeeded.
 */
__must_check int afs_event(enum afs_events event, struct para_buffer *pb,
		void *data)
{
	int i, ret;

	for (i = 0; i < NUM_AFS_TABLES; i++) {
		const struct afs_table *t = afs_tables + i;
		if (!t->ops->event_handler)
			continue;
		ret = t->ops->event_handler(event, pb, data);
		if (ret < 0) {
			PARA_CRIT_LOG("table %s, event %u: %s\n", t->name,
				event, para_strerror(-ret));
			return ret;
		}
	}
	return 1;
}
