/* Copyright (C) 2007 Andre Noll <maan@tuebingen.mpg.de>, see file COPYING. */

/** \file afs.h Exported symbols of the audio file selector. */

/** Audio file selector data stored in the audio file table. */
struct afs_info {
	/** Seconds since the epoch. */
	uint64_t last_played;
	/** Bit field of set attributes. */
	uint64_t attributes;
	/** Counts how many times the file was selected. */
	uint32_t num_played;
	/** Image blob associated with this file (foreign key). */
	uint32_t image_id;
	/** Lyrics blob associated with this file (foreign key). */
	uint32_t lyrics_id;
	/** Mp3, ogg, ... */
	uint8_t audio_format_id;
	/** Amplification value. */
	uint8_t amp;
};

/**
 * Events caused by changes to an afs table.
 *
 * Whenever an afs table changes, an event is generated which causes afs to
 * call the event handlers of all other tables. For example, if an audio file
 * is added, the event handler of the mood table checks the new file for
 * admissibility.
 */
enum afs_events {
	/** An attribute was added. */
	ATTRIBUTE_ADD,
	/** An attribute was renamed. */
	ATTRIBUTE_RENAME,
	/** An attribute was removed. */
	ATTRIBUTE_REMOVE,
	/** The afs info struct of an audio file changed. */
	AFSI_CHANGE,
	/** The afh info struct of an audio file changed. */
	AFHI_CHANGE,
	/** The vss streams a new file, implies afsi and afhi change */
	VSS_NEW_AUDIO_FILE,
	/** An audio file was renamed. */
	AUDIO_FILE_RENAME,
	/** An audio file was added. */
	AUDIO_FILE_ADD,
	/** An audio file is about to be removed. */
	AUDIO_FILE_REMOVE,
	/** A new blob was added. */
	BLOB_ADD,
	/** A blob was renamed. */
	BLOB_RENAME,
	/** A blob is about to be removed. */
	BLOB_REMOVE,
};

/**
 * Used as data for \ref afs_event() for events of type \p ATTRIBUTE_ADD.
 */
struct rmatt_event_data {
	/** The bit number of the attribute being removed. */
	unsigned char bitnum;
};

/**
 * Used as data for \ref afs_event() for events of type \p ATTRIBUTE_AFSI_CHANGE.
 */
struct afsi_change_event_data {
	/** Pointer to the row that has changed. */
	struct osl_row *aft_row;
	/** Afs info before the change. */
	struct afs_info *old_afsi;
};

/** Methods for table startup/shutdown and event handling. */
struct afs_table_operations {
	/** Gets called on startup and on SIGHUP. */
	int (*open)(const char *base_dir);
	/** Gets called on shutdown and on SIGHUP. */
	void (*close)(void);
	/** Called from the init command. */
	int (*create)(const char *);
	/** Handle events generated by other tables. See enum \ref afs_events. */
	int (*event_handler)(enum afs_events event, struct para_buffer *pb,
		void *data);
};

/** How audio files are selected by afs. */
enum play_mode {
	/** Admissible files are determined by a mood definition. */
	PLAY_MODE_MOOD,
	/** All listed files are admissible. */
	PLAY_MODE_PLAYLIST,
};

/**
 * Codes used for communication between the server and the afs process.
 *
 * Before forking the afs child, para_server creates a bidirectional pipe
 * through which both processes communicate. Usually para_server requests a new
 * audio file in order to start streaming or when the end of the current audio file
 * has been reached.  The afs process responds to such a request by sending
 * back an eight byte buffer. The first four bytes is the uint32_t
 * representation of the code, usually \p NEXT_AUDIO_FILE if an admissible
 * audio file was found, successfully opened and verified. The other four bytes
 * represent the shared memory id of the shared memory area that contains
 * details about the audio file to be streamed next. The open file descriptor
 * of that file is also passed from afs to para_server through the same pipe.
 */
enum afs_server_code {
	/** An audio file was successfully opened. */
	NEXT_AUDIO_FILE,
	/** No admissible audio file was found. */
	NO_ADMISSIBLE_FILES,
};

/** Flags passed to for_each_matching_row(). */
enum pattern_match_flags {
	/** Loop in reverse order. */
	PM_REVERSE_LOOP = 1,
	/** If no pattern is given, loop over all rows. */
	PM_NO_PATTERN_MATCHES_EVERYTHING = 2,
	/** If the data in match_column is the empty string, skip this row. */
	PM_SKIP_EMPTY_NAME = 4,
};

/** Structure passed to for_each_matching_row(). */
struct pattern_match_data {
	/** Loop over all rows in this table. */
	struct osl_table *table;
	/** Determines the loop order. Must be an rbtree column. */
	unsigned loop_col_num;
	/** Data from this column is matched against the given patterns. */
	unsigned match_col_num;
	/** \see \ref pattern_match_flags. */
	unsigned pm_flags;
	/** This value is passed verbatim to fnmatch(). */
	int fnmatch_flags;
	/** Obtained by deserializing the query buffer in the callback. */
	struct lls_parse_result *lpr;
	/** Do not try to match the first inputs of lpr */
	unsigned input_skip;
	/** Data pointer passed to the action function. */
	void *data;
	/** Gets increased by one for each match. */
	unsigned num_matches;
	/** For each matching row, this function will be called. */
	int (*action)(struct osl_table *table, struct osl_row *row, const char *name, void *data);
};

/** Arguments passed to each afs callback. */
struct afs_callback_arg {
	/** The local socket connecting afs and the command handler. */
	int fd;
	/** Callback-specific data. */
	struct osl_object query;
	/** Will be written on band SBD_OUTPUT, fully buffered. */
	struct para_buffer pbout;
	/**
	 * Convenience pointer for the deserialized parse result.
	 *
	 * Most afs command handlers call \ref send_lls_callback_request() to
	 * serialize the parse result of the subcommand and pass it to the
	 * callback. In afs context a pointer to the deserialized parse result
	 * is stored here.
	 */
	struct lls_parse_result *lpr;
};

/**
 * The "top half" of an afs command.
 *
 * Afs command handlers run as a process which is not related to the afs
 * process, i.e. they can not change the address space of afs directly.
 * Therefore afs commands typically consist of two functions: The command
 * handler and the corresponding callback function that runs in afs context.
 *
 * \sa \ref send_callback_request().
 */
typedef int afs_callback(struct afs_callback_arg *aca);

/**
 * Dispatch the output of an afs callback.
 *
 * Some AFS callbacks need to send data back to the command handler. Pointers
 * to this type of function are passed to \ref send_callback_request() and
 * related functions to dispatch the data in the command handler process. Most
 * (but not all) afs commands pass \ref afs_cb_result_handler(), which sends
 * the output of the callback to the connected client.
 */
typedef int callback_result_handler(struct osl_object *result, uint8_t band, void *private);
int afs_cb_result_handler(struct osl_object *result, uint8_t band, void *private);
int pass_buffer_as_shm(int fd, uint8_t band, const char *buf, size_t size);

/** Structure passed to the AFS max_size handler. */
struct afs_max_size_handler_data {
	/** Local socket connecting the command handler and the AFS process. */
	int fd;
	/** The sideband designator for this data packet. */
	uint8_t band;
};

/**
 * Standard max_size handler for AFS commands.
 *
 * \param buf Contains (part of) the AFS command output.
 * \param size The number of bytes in \a buf.
 * \param private Pointer to a \ref afs_max_size_handler_data structure.
 *
 * Whenever the output of an AFS command exceeds the maximal size of a shared
 * memory area, the max size handler of the para_buffer which holds the command
 * output is called with \a private being a pointer to a structure of type
 * afs_max_size_handler_data.
 *
 * \return The return value of the underlying call to \ref
 *  pass_buffer_as_shm().
 */
_static_inline_ int afs_max_size_handler(char *buf, size_t size, void *private)
{
	struct afs_max_size_handler_data *amshd = private;
	return pass_buffer_as_shm(amshd->fd, amshd->band, buf, size);
}

__noreturn void afs_init(int socket_fd);
__must_check int afs_event(enum afs_events event, struct para_buffer *pb,
	void *data);
int send_callback_request(afs_callback *f, struct osl_object *query,
		callback_result_handler *result_handler,
		void *private_result_data);
int send_lls_callback_request(afs_callback *f,
		const struct lls_command * const cmd,
		struct lls_parse_result *lpr, void *private_result_data);
__printf_2_3 void afs_error(const struct afs_callback_arg *aca,
		const char *fmt,...);
int string_compare(const struct osl_object *obj1, const struct osl_object *obj2);
int for_each_matching_row(struct pattern_match_data *pmd);

/* score */
extern const struct afs_table_operations score_ops;
void score_open(struct osl_table **result);
void score_close(struct osl_table *t);
int score_loop(osl_rbtree_loop_func *func, struct osl_table *t, void *data);
int score_get_best(struct osl_row **aft_row, long *score);
int get_score_and_aft_row(struct osl_row *score_row, long *score, struct osl_row **aft_row);
int score_add(const struct osl_row *aft_row, long score, struct osl_table *t);
int score_update(const struct osl_row *aft_row, long new_score);
int score_delete(const struct osl_row *aft_row);
void score_clear(void);
bool row_belongs_to_score_table(const struct osl_row *aft_row);

/* attribute */
extern const struct afs_table_operations attr_ops;
void get_attribute_bitmap(const uint64_t *atts, char *buf); /* needed by com_ls() */
int get_attribute_bitnum_by_name(const char *att_name, unsigned char *bitnum);
int get_attribute_text(uint64_t *atts, const char *delim, char **text);
int attribute_check_callback(struct afs_callback_arg *aca);

/* aft */
extern const struct afs_table_operations aft_ops;
int aft_get_row_of_path(const char *path, struct osl_row **row);
int aft_check_attributes(uint64_t att_mask, struct para_buffer *pb);
int open_and_update_audio_file(int *fd);
int load_afd(int shmid, struct audio_file_data *afd);
int get_afsi_of_row(const struct osl_row *row, struct afs_info *afsi);
int get_afhi_of_row(const struct osl_row *row, struct afh_info *afhi);
int get_audio_file_path_of_row(const struct osl_row *row, char **path);
int audio_file_loop(void *private_data, osl_rbtree_loop_func *func);
int aft_check_callback(struct afs_callback_arg *aca);
void free_status_items(void);

/* mood */
struct mood_instance;
int mood_load(const char *mood_name, struct mood_instance **result, char **msg);
int mood_loop(struct mood_instance *m, osl_rbtree_loop_func *func, void *data);
void mood_unload(struct mood_instance *m);
int mood_check_callback(struct afs_callback_arg *aca);

/* playlist */
struct playlist_instance;
int playlist_load(const char *name, struct playlist_instance **result, char **msg);
int playlist_loop(struct playlist_instance *pi, osl_rbtree_loop_func *func, void *data);
void playlist_unload(struct playlist_instance *pi);
int playlist_check_callback(struct afs_callback_arg *aca);

/** evaluates to 1 if x < y, to -1 if x > y and to 0 if x == y */
#define NUM_COMPARE(x, y) ((int)((x) < (y)) - (int)((x) > (y)))


/** Define exported functions and a table pointer for an osl blob table. */
#define DECLARE_BLOB_SYMBOLS(table_name, cmd_prefix) \
	int cmd_prefix ## _get_name_by_id(uint32_t id, char **name); \
	int cmd_prefix ## _get_def_by_id(uint32_t id, struct osl_object *def); \
	int cmd_prefix ## _get_def_by_name(const char *name, struct osl_object *def); \
	int cmd_prefix ## _get_name_and_def_by_row(const struct osl_row *row, \
		char **name, struct osl_object *def); \
	int table_name ##_event_handler(enum afs_events event, \
		struct para_buffer *pb, void *data); \
	extern struct osl_table *table_name ## _table; \
	extern const struct afs_table_operations table_name ## _ops;

/** \cond blob_symbols */
DECLARE_BLOB_SYMBOLS(lyrics, lyr)
DECLARE_BLOB_SYMBOLS(images, img)
DECLARE_BLOB_SYMBOLS(moods, mood)
DECLARE_BLOB_SYMBOLS(playlists, pl)
/** \endcond blob_symbols */

/** The columns of an abstract blob table. */
enum blob_table_columns {
	/** The identifier, a positive integer that never repeats. */
	BLOBCOL_ID,
	/** The unique name of the blob. */
	BLOBCOL_NAME,
	/** The actual blob contents. */
	BLOBCOL_DEF,
	/** A blob table has that many columns. */
	NUM_BLOB_COLUMNS
};
