/* Copyright (C) 1997 Andre Noll <maan@tuebingen.mpg.de>, see file COPYING. */

/** \file attribute.c Attribute handling functions. */

#include <osl.h>
#include <lopsub.h>

#include "server_cmd.lsg.h"
#include "para.h"
#include "error.h"
#include "crypt.h"
#include "string.h"
#include "afh.h"
#include "afs.h"
#include "ipc.h"
#include "sideband.h"
#include "command.h"

static struct osl_table *attribute_table;
static int greatest_att_bitnum;

/** The columns of the attribute table. */
enum attribute_table_columns {
	/** The bit number (0-63). */
	ATTCOL_BITNUM,
	/** The name of the attribute. */
	ATTCOL_NAME,
	/** Number of columns in this table. */
	NUM_ATT_COLUMNS
};

static int char_compare(const struct osl_object *obj1, const struct osl_object *obj2)
{
	const unsigned char *c1 = (const unsigned char*)obj1->data;
	const unsigned char *c2 = (const unsigned char*)obj2->data;
	if (*c1 > *c2)
		return 1;
	if (*c1 < *c2)
		return -1;
	return 0;
}

static struct osl_column_description att_cols[] = {
	[ATTCOL_BITNUM] = {
		.storage_type = OSL_MAPPED_STORAGE,
		.storage_flags = OSL_RBTREE | OSL_FIXED_SIZE | OSL_UNIQUE,
		.name = "bitnum",
		.compare_function = char_compare,
		.data_size = 1
	},
	[ATTCOL_NAME] = {
		.storage_type = OSL_MAPPED_STORAGE,
		.storage_flags = OSL_RBTREE | OSL_UNIQUE,
		.name = "name",
		.compare_function = string_compare,
	}
};

static struct osl_table_description attribute_table_desc = {
	.name = "attributes",
	.num_columns = NUM_ATT_COLUMNS,
	.flags = 0,
	.column_descriptions = att_cols
};

static void find_greatest_att_bitnum(void)
{
	unsigned char c = 63;
	do {
		struct osl_row *row;
		struct osl_object obj = {.data = &c, .size = 1};
		if (osl_get_row(attribute_table, ATTCOL_BITNUM, &obj,
				&row) >= 0) {
			greatest_att_bitnum = c;
			return;
		}
	} while (c--);
	PARA_INFO_LOG("no attributes\n");
	greatest_att_bitnum = -E_NO_ATTRIBUTES;
}

/**
 * Retrieve the identifier (number) of an attribute.
 *
 * \param att_name The name of the attribute.
 * \param bitnum Result pointer.
 *
 * \return Positive on success, negative on errors.
 */
int get_attribute_bitnum_by_name(const char *att_name, unsigned char *bitnum)
{
	struct osl_object obj = {.data = (char *)att_name,
		.size = strlen(att_name) + 1};
	struct osl_row *row;
	int ret = osl(osl_get_row(attribute_table, ATTCOL_NAME, &obj, &row));

	if (ret < 0)
		return ret;
	ret = osl(osl_get_object(attribute_table, row, ATTCOL_BITNUM, &obj));
	if (ret < 0)
		return ret;
	*bitnum = *(unsigned char *)obj.data;
	return 1;
}

/* Data passed to the action function of lsatt */
static int print_attribute(struct osl_table *table, struct osl_row *row,
		const char *name, void *data)
{
	struct afs_callback_arg *aca = data;
	bool l_given = SERVER_CMD_OPT_GIVEN(LSATT, LONG, aca->lpr);
	struct osl_object bitnum_obj;
	int ret;

	if (!l_given) {
		para_printf(&aca->pbout, "%s\n", name);
		return 1;
	}
	ret = osl(osl_get_object(table, row, ATTCOL_BITNUM, &bitnum_obj));
	if (ret < 0) {
		afs_error(aca, "%s: %s\n", name, para_strerror(-ret));
		return ret;
	}
	para_printf(&aca->pbout, "%u\t%s\n", *(unsigned char*)bitnum_obj.data,
		name);
	return 1;
}

static int com_lsatt_callback(struct afs_callback_arg *aca)
{
	const struct lls_command *cmd = SERVER_CMD_CMD_PTR(LSATT);
	bool i_given, r_given;
	int ret;
	struct pattern_match_data pmd = {
		.table = attribute_table,
		.loop_col_num = ATTCOL_NAME,
		.match_col_num = ATTCOL_NAME,
		.pm_flags = PM_NO_PATTERN_MATCHES_EVERYTHING,
		.data = aca,
		.action = print_attribute
	};

	ret = lls(lls_deserialize_parse_result(aca->query.data, cmd, &aca->lpr));
	assert(ret >= 0);
	pmd.lpr = aca->lpr;
	i_given = SERVER_CMD_OPT_GIVEN(LSATT, ID_SORT, aca->lpr);
	r_given = SERVER_CMD_OPT_GIVEN(LSATT, REVERSE, aca->lpr);

	if (i_given)
		pmd.loop_col_num = ATTCOL_BITNUM;
	if (r_given)
		pmd.pm_flags |= PM_REVERSE_LOOP;
	ret = for_each_matching_row(&pmd);
	if (ret < 0)
		goto out;
	if (pmd.num_matches == 0)
		ret = -E_NO_MATCH;
out:
	lls_free_parse_result(aca->lpr, cmd);
	return ret;
}

static int com_lsatt(struct command_context *cc, struct lls_parse_result *lpr)
{
	const struct lls_command *cmd = SERVER_CMD_CMD_PTR(LSATT);
	return send_lls_callback_request(com_lsatt_callback, cmd, lpr, cc);
}
EXPORT_SERVER_CMD_HANDLER(lsatt);

static int com_addatt_callback(struct afs_callback_arg *aca)
{
	const struct lls_command *cmd = SERVER_CMD_CMD_PTR(ADDATT);
	int i, ret = 1;
	size_t len;
	unsigned num_inputs;

	ret = lls(lls_deserialize_parse_result(aca->query.data, cmd, &aca->lpr));
	assert(ret >= 0);
	num_inputs = lls_num_inputs(aca->lpr);
	for (i = 0; i < num_inputs; i++) {
		const char *name = lls_input(i, aca->lpr);
		struct osl_object objs[NUM_ATT_COLUMNS];
		struct osl_row *row;
		unsigned char bitnum;

		len = strlen(name);
		if (len == 0 || name[len - 1] == '-' || name[len - 1] == '+') {
			afs_error(aca, "invalid attribute name: %s\n", name);
			continue;
		}
		ret = get_attribute_bitnum_by_name(name, &bitnum);
		if (ret >= 0) {
			para_printf(&aca->pbout,
				"attribute \"%s\" already exists\n", name);
			continue;
		}
		if (ret != -OSL_ERRNO_TO_PARA_ERROR(E_OSL_RB_KEY_NOT_FOUND)) /* error */
			goto out;
		objs[ATTCOL_BITNUM].size = 1;
		/* find smallest unused attribute */
		for (bitnum = 0; bitnum < 64; bitnum++) {
			objs[ATTCOL_BITNUM].data = &bitnum;
			ret = osl(osl_get_row(attribute_table, ATTCOL_BITNUM,
				&objs[ATTCOL_BITNUM], &row));
			if (ret == -OSL_ERRNO_TO_PARA_ERROR(E_OSL_RB_KEY_NOT_FOUND))
				break; /* this bitnum is unused, use it */
			if (ret < 0) /* error */
				goto out;
			/* this bit is already in use, try next bit */
		}
		if (bitnum == 64) {
			ret = -E_ATT_TABLE_FULL;
			goto out;
		}
		objs[ATTCOL_NAME].data = (char *)name;
		objs[ATTCOL_NAME].size = len + 1;
		ret = osl(osl_add_row(attribute_table, objs));
		if (ret < 0)
			goto out;
		ret = afs_event(ATTRIBUTE_ADD, &aca->pbout, NULL);
		if (ret < 0)
			goto out;
		greatest_att_bitnum = PARA_MAX(greatest_att_bitnum, (int)bitnum);
	}
out:
	if (ret < 0)
		afs_error(aca, "error while adding %s\n",
			lls_input(i, aca->lpr));
	lls_free_parse_result(aca->lpr, cmd);
	return ret;
}

static int com_addatt(struct command_context *cc, struct lls_parse_result *lpr)
{
	const struct lls_command *cmd = SERVER_CMD_CMD_PTR(ADDATT);
	char *errctx;
	int ret = lls(lls_check_arg_count(lpr, 1, 64, &errctx));

	if (ret < 0) {
		send_errctx(cc, errctx);
		return ret;
	}
	return send_lls_callback_request(com_addatt_callback, cmd, lpr, cc);
}
EXPORT_SERVER_CMD_HANDLER(addatt);

static int com_mvatt_callback(struct afs_callback_arg *aca)
{
	const struct lls_command *cmd = SERVER_CMD_CMD_PTR(MVATT);
	const char *old, *new;
	struct osl_object obj;
	struct osl_row *row;
	int ret;

	ret = lls(lls_deserialize_parse_result(aca->query.data, cmd, &aca->lpr));
	assert(ret >= 0);
	old = lls_input(0, aca->lpr);
	new = lls_input(1, aca->lpr);
	obj.data = (char *)old;
	obj.size = strlen(old) + 1;
	ret = osl(osl_get_row(attribute_table, ATTCOL_NAME, &obj, &row));
	if (ret < 0)
		goto out;
	obj.data = (char *)new;
	obj.size = strlen(new) + 1;
	/* The update fails if the destination attribute exists. */
	ret = osl(osl_update_object(attribute_table, row, ATTCOL_NAME, &obj));
out:
	if (ret < 0)
		afs_error(aca, "cannot rename %s to %s\n", old, new);
	else
		ret = afs_event(ATTRIBUTE_RENAME, &aca->pbout, NULL);
	lls_free_parse_result(aca->lpr, cmd);
	return ret;
}

static int com_mvatt(struct command_context *cc, struct lls_parse_result *lpr)
{
	const struct lls_command *cmd = SERVER_CMD_CMD_PTR(MVATT);
	char *errctx;
	int ret = lls(lls_check_arg_count(lpr, 2, 2, &errctx));
	if (ret < 0) {
		send_errctx(cc, errctx);
		return ret;
	}
	return send_lls_callback_request(com_mvatt_callback, cmd, lpr, cc);
}
EXPORT_SERVER_CMD_HANDLER(mvatt);

static int remove_attribute(struct osl_table *table, struct osl_row *row,
		const char *name, void *data)
{
	struct afs_callback_arg *aca = data;
	int ret;
	struct rmatt_event_data red;

	ret = get_attribute_bitnum_by_name(name, &red.bitnum);
	if (ret < 0) {
		afs_error(aca, "cannot remove %s\n", name);
		return ret;
	}
	para_printf(&aca->pbout, "removing attribute %s\n", name);
	ret = osl(osl_del_row(table, row));
	if (ret < 0) {
		afs_error(aca, "cannot remove %s\n", name);
		return ret;
	}
	return afs_event(ATTRIBUTE_REMOVE, &aca->pbout, &red);
}

static int com_rmatt_callback(struct afs_callback_arg *aca)
{
	const struct lls_command *cmd = SERVER_CMD_CMD_PTR(RMATT);
	int ret;
	struct pattern_match_data pmd = {
		.table = attribute_table,
		.loop_col_num = ATTCOL_BITNUM,
		.match_col_num = ATTCOL_NAME,
		.data = aca,
		.action = remove_attribute
	};
	ret = lls(lls_deserialize_parse_result(aca->query.data, cmd, &aca->lpr));
	assert(ret >= 0);
	pmd.lpr = aca->lpr;
	ret = for_each_matching_row(&pmd);
	if (ret < 0)
		goto out;
	if (pmd.num_matches == 0)
		ret = -E_NO_MATCH;
out:
	lls_free_parse_result(aca->lpr, cmd);
	return ret;
}

static int com_rmatt(struct command_context *cc, struct lls_parse_result *lpr)
{
	const struct lls_command *cmd = SERVER_CMD_CMD_PTR(RMATT);
	char *errctx;
	int ret = lls(lls_check_arg_count(lpr, 1, INT_MAX, &errctx));

	if (ret < 0) {
		send_errctx(cc, errctx);
		return ret;
	}
	return send_lls_callback_request(com_rmatt_callback, cmd, lpr, cc);
}
EXPORT_SERVER_CMD_HANDLER(rmatt);

/**
 * Return a binary representation of the given attribute value.
 *
 * \param atts Pointer to the attribute value.
 * \param buf Result.
 *
 * This function prints a string of at most 64 characters plus the terminating
 * \p NULL character into \a buf which must be provided by the caller and at
 * least 65 bytes long. The "x" character is used for set attributes and "-" is
 * used for unset attributes.
 *
 * In practice, not all 64 attributes are defined. In this case, the function
 * only prints \a N + 1 characters where \a N is the greatest id of a defined
 * attribute.
 */
void get_attribute_bitmap(const uint64_t *atts, char *buf)
{
	int i;
	const uint64_t one = 1;

	for (i = 0; i <= greatest_att_bitnum; i++)
		buf[greatest_att_bitnum - i] = (*atts & (one << i))? 'x' : '-';
	buf[i] = '\0';
}

/**
 * Get a string containing the set attributes in text form.
 *
 * \param atts The attribute bitmap.
 * \param delim The delimiter to separate matching attribute names.
 * \param text Result pointer.
 *
 * \return Positive on success, negative on errors. If no attributes have
 * been defined, \a *text is NULL.
 */
int get_attribute_text(uint64_t *atts, const char *delim, char **text)
{
	int i, ret;
	const uint64_t one = 1;

	*text = NULL;
	if (greatest_att_bitnum < 0) { /* no attributes available */
		*text = para_strdup("(no attributes available)");
		return 1;
	}
	for (i = 0; i <= greatest_att_bitnum; i++) {
		unsigned char bn = i;
		struct osl_object obj = {.data = &bn, .size = 1};
		struct osl_row *row;

		if (!(*atts & (one << i)))
			continue;
		ret = osl(osl_get_row(attribute_table, ATTCOL_BITNUM, &obj, &row));
		if (ret < 0)
			goto err;
		ret = osl(osl_get_object(attribute_table, row, ATTCOL_NAME, &obj));
		if (ret < 0)
			goto err;
		if (*text) {
			char *tmp = make_message("%s%s%s", *text, delim, (char *)obj.data);
			free(*text);
			*text = tmp;
		} else
			*text = para_strdup(obj.data);
	}
	if (!*text) /* no attributes set */
		*text = para_strdup("");
	return 1;
err:
	free(*text);
	return ret;
}

static int att_logical_or(struct osl_row *row, void *data)
{
	uint64_t *att_mask = data, one = 1;
	struct osl_object bitnum_obj;
	int ret = osl_get_object(attribute_table, row, ATTCOL_BITNUM, &bitnum_obj);

	if (ret < 0)
		return ret;
	*att_mask |= one << *(unsigned char *)bitnum_obj.data;
	return 0;
}

/**
 * Compute the attribute bit mask and check each afs info bitmap.
 *
 * \param aca The query field of \a aca is ignored.
 *
 * This iterates over all attributes in the attribute table and computes the
 * logical or of 1 << b where b is the bit number of the attribute. The
 * resulting bit mask is passed to aft_check_attributes() which performs the
 * actual check.
 *
 * \return Standard.
 *
 * \sa \ref aft_check_attributes().
 */
int attribute_check_callback(struct afs_callback_arg *aca)
{
	int ret;
	uint64_t att_mask = 0; /* bits corresponding to a attributes */

	ret = osl_rbtree_loop(attribute_table, ATTCOL_BITNUM, &att_mask,
		att_logical_or);
	if (ret < 0) {
		PARA_ERROR_LOG("attribute table loop failed: %s\n",
			para_strerror(-ret));
		return ret;
	}
	return aft_check_attributes(att_mask, &aca->pbout);
}

static void attribute_close(void)
{
	osl_close_table(attribute_table, OSL_MARK_CLEAN);
	attribute_table = NULL;
}

static int attribute_open(const char *dir)
{
	int ret;

	attribute_table_desc.dir = dir;
	ret = osl(osl_open_table(&attribute_table_desc, &attribute_table));
	greatest_att_bitnum = -1; /* no atts available */
	if (ret >= 0) {
		find_greatest_att_bitnum();
		return ret;
	}
	attribute_table = NULL;
	if (ret == -OSL_ERRNO_TO_PARA_ERROR(E_OSL_NOENT))
		return 1;
	return ret;
}

static int attribute_create(const char *dir)
{
	attribute_table_desc.dir = dir;
	return osl(osl_create_table(&attribute_table_desc));
}

/** The attribute table stores name/bitnum pairs. */
const struct afs_table_operations attr_ops = { /* no event handler */
	.open = attribute_open,
	.close = attribute_close,
	.create = attribute_create,
};
