/* Copyright (C) 2006 Andre Noll <maan@tuebingen.mpg.de>, see file COPYING. */

/** \file dccp_send.c Paraslash's dccp sender. */

/*
 * based on server.c of dccp-cs-0.01.tar.bz2,
 * (C) 2005 Ian McDonald <imcdnzl@gmail.com>
 */

#include <netinet/in.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <arpa/inet.h>
#include <sys/un.h>
#include <netdb.h>
#include <lopsub.h>

#include "server.lsg.h"
#include "para.h"
#include "error.h"
#include "string.h"
#include "afh.h"
#include "net.h"
#include "server.h"
#include "list.h"
#include "sched.h"
#include "send.h"
#include "vss.h"
#include "fd.h"

static struct sender_status dccp_sender_status, *dss = &dccp_sender_status;

struct dccp_fec_client {
	struct fec_client_parms fcp;
	struct fec_client *fc;
};

static void dccp_pre_monitor(struct sched *s)
{
	unsigned n;

	FOR_EACH_LISTEN_FD(n, dss)
		if (dss->listen_fds[n] >= 0)
			sched_monitor_readfd(dss->listen_fds[n], s);
}

#ifndef DCCP_SOCKOPT_TX_CCID
#define DCCP_SOCKOPT_TX_CCID 14 /**< Set/get the TX CCID. */
#endif

/**
 * Query the TX CCID used on the sender-client half connection.
 * \param sockfd Server socket descriptor to query (after accept(2)).
 * \return CCID number > 0, -1 on error/incompatibility.
 *
 * NB: This feature is only available on Linux > 2.6.30; on older kernels
 * ENOPROTOOPT ("Protocol not available") will be returned.
 */
static int dccp_get_tx_ccid(int sockfd)
{
	int tx_ccid;
	socklen_t len = sizeof(tx_ccid);

	if (getsockopt(sockfd, SOL_DCCP,
		       DCCP_SOCKOPT_TX_CCID, &tx_ccid, &len) < 0) {
		PARA_WARNING_LOG("DCCP_SOCKOPT_TX_CCID: %s\n", strerror(errno));
		return -1;
	}
	return tx_ccid;
}

static void dccp_shutdown_client(struct sender_client *sc)
{
	struct dccp_fec_client *dfc = sc->private_data;

	vss_del_fec_client(dfc->fc);
	shutdown_client(sc, dss);
}

static void dccp_shutdown_clients(void)
{
	struct sender_client *sc, *tmp;

	list_for_each_entry_safe(sc, tmp, &dss->client_list, node)
		dccp_shutdown_client(sc);
}

static void dccp_shutdown(void)
{
	dccp_shutdown_clients();
	generic_acl_deplete(&dss->acl);
	free_sender_status(dss);
}

#ifndef DCCP_SOCKOPT_GET_CUR_MPS
#define DCCP_SOCKOPT_GET_CUR_MPS 5 /**< Max packet size, RFC 4340, 14. */
#endif

/** Estimated worst-case length of a DCCP header including options. */
#define DCCP_MAX_HEADER 128

/** * Obtain current MPS according to RFC 4340, sec. 14. */
static int dccp_init_fec(struct sender_client *sc)
{
	int mps, ret;
	socklen_t ml = sizeof(mps);
	uint32_t mss; /* max slize size */

	/* If call fails, return some sensible minimum value */
	ret = getsockopt(sc->fd, SOL_DCCP, DCCP_SOCKOPT_GET_CUR_MPS, &mps, &ml);
	if (ret < 0) {
		PARA_NOTICE_LOG("can not determine MPS: %s\n", strerror(errno));
		mps = generic_max_transport_msg_size(sc->fd) - DCCP_MAX_HEADER;
	}
	PARA_INFO_LOG("current MPS = %d bytes\n", mps);
	assert(mps > 0);
	mss = OPT_UINT32_VAL(DCCP_MAX_SLICE_SIZE);
	if (mss > 0 && mss <= INT_MAX)
		mps = PARA_MIN(mps, (int)mss);
	return mps;
}

static void dccp_send_fec(struct sender_client *sc, char *buf, size_t len)
{
	int ret = xwrite(sc->fd, buf, len);

	if (ret < 0)
		dccp_shutdown_client(sc);
}

static void dccp_post_monitor(__a_unused struct sched *s)
{
	struct sender_client *sc;
	struct dccp_fec_client *dfc;
	int tx_ccid;
	uint32_t k, n;

	sc = accept_sender_client(dss);
	if (!sc)
		return;

	/* If CCID identifiable, present client as <host>#<port>~<ccid> */
	tx_ccid = dccp_get_tx_ccid(sc->fd);
	if (tx_ccid != -1) {
		char *tmp = make_message("%s~%d", sc->name, tx_ccid);

		free(sc->name);
		sc->name = tmp;
	}
	/*
	 * Bypass unused CCID paths: the sender does not receive application data
	 * from the client; by shutting down this unused communication path we can
	 * reduce processing costs a bit. See analogous comment in \ref dccp_recv.c.
	 */
	if (shutdown(sc->fd, SHUT_RD) < 0) {
		PARA_WARNING_LOG("%s\n", strerror(errno));
		shutdown_client(sc, dss);
		return;
	}
	dfc = zalloc(sizeof(*dfc));
	sc->private_data = dfc;
	k = OPT_UINT32_VAL(DCCP_DATA_SLICES_PER_GROUP);
	n = OPT_UINT32_VAL(DCCP_SLICES_PER_GROUP);
	if (k == 0 || n == 0 || k >= n) {
		PARA_WARNING_LOG("invalid FEC parameters, using defaults\n");
		dfc->fcp.data_slices_per_group = 3;
		dfc->fcp.slices_per_group = 4;
	} else {
		dfc->fcp.data_slices_per_group = k;
		dfc->fcp.slices_per_group = n;
	}
	dfc->fcp.init_fec		= dccp_init_fec;
	dfc->fcp.send_fec		= dccp_send_fec;
	dfc->fcp.need_periodic_header   = false;
	dfc->fc = vss_add_fec_client(sc, &dfc->fcp);
}

static int dccp_com_on(__a_unused struct sender_command_data *scd)
{
	generic_com_on(dss, IPPROTO_DCCP);
	return 1;
}

static int dccp_com_off(__a_unused struct sender_command_data *scd)
{
	dccp_shutdown_clients();
	generic_com_off(dss);
	return 1;
}


static int dccp_com_deny(struct sender_command_data *scd)
{
	generic_com_deny(scd, dss);
	return 1;
}

static int dccp_com_allow(struct sender_command_data *scd)
{
	generic_com_allow(scd, dss);
	return 1;
}

/**
 * Return list of available CCIDs or warning, in static buffer.
 */
static const char *dccp_list_available_ccids(void)
{
	/* Worst case length: n * 3 digits + n-1 spaces + '\0' */
	static char list[DCCP_MAX_HOST_CCIDS * 4];
	uint8_t *ccids;
	int i, len, nccids;

	nccids = dccp_available_ccids(&ccids);
	if (nccids < 0) {
		snprintf(list, sizeof(list), "Unable to query available CCIDs");
	} else {
		for (i = len = 0; i < nccids; i++)
			len += snprintf(list + len, sizeof(list) - len,
					"%s%d", i ? " " : "", ccids[i]);
	}
	return list;
}

static char *dccp_status(void)
{
	char *status = generic_sender_status(dss, "dccp");
	char *result = make_message("%ssupported ccids: %s\n", status,
		dccp_list_available_ccids());
	free(status);
	return result;
}

/*
 * Initialize the client list and the access control list and listen on the
 * dccp port.
 */
static void dccp_send_init(void)
{
	init_sender_status(dss, OPT_RESULT(DCCP_ACCESS),
		OPT_RESULT(DCCP_LISTEN_ADDRESS),
		OPT_UINT32_VAL(DCCP_PORT), OPT_UINT32_VAL(DCCP_MAX_CLIENTS),
		OPT_GIVEN(DCCP_DEFAULT_DENY));
	if (OPT_GIVEN(DCCP_NO_AUTOSTART))
		return;
	generic_com_on(dss, IPPROTO_DCCP);
}

/**
 * The DCCP sender.
 *
 * This sender offers congestion control not available in plain TCP. Most
 * methods of the sender structure are implemented as simple wrappers for the
 * generic functions defined in \ref send_common.c. Like UDP streams, DCCP
 * streams are sent FEC-encoded.
 */
const struct sender dccp_sender = {
	.name = "dccp",
	.init = dccp_send_init,
	.shutdown = dccp_shutdown,
	.pre_monitor = dccp_pre_monitor,
	.post_monitor = dccp_post_monitor,
	.shutdown_clients = dccp_shutdown_clients,
	.client_cmds = {
		[SENDER_on] = dccp_com_on,
		[SENDER_off] = dccp_com_off,
		[SENDER_deny] = dccp_com_deny,
		[SENDER_allow] = dccp_com_allow,
	},
	.help = generic_sender_help,
	.status = dccp_status,
};
