/* Copyright (C) 2006 Andre Noll <maan@tuebingen.mpg.de>, see file COPYING. */

/** \file fd.c Helper functions for file descriptor handling. */

#include <sys/types.h>
#include <dirent.h>
#include <sys/mman.h>

#include "para.h"
#include "error.h"
#include "string.h"
#include "fd.h"

/**
 * Change the name or location of a file.
 *
 * \param oldpath File to be moved.
 * \param newpath Destination.
 *
 * This is just a simple wrapper for the rename(2) system call which returns a
 * paraslash error code and prints an error message on failure.
 *
 * \return Standard.
 *
 * \sa rename(2).
 */
int xrename(const char *oldpath, const char *newpath)
{
	int ret = rename(oldpath, newpath);

	if (ret >= 0)
		return 1;
	ret = -ERRNO_TO_PARA_ERROR(errno);
	PARA_ERROR_LOG("failed to rename %s -> %s\n", oldpath, newpath);
	return ret;
}

/**
 * Write an array of buffers, handling non-fatal errors.
 *
 * \param fd The file descriptor to write to.
 * \param iov Pointer to one or more buffers.
 * \param iovcnt The number of buffers.
 *
 * EAGAIN, EWOULDBLOCK and EINTR are not considered error conditions. If a
 * write operation fails with EAGAIN or EWOULDBLOCK, the number of bytes that
 * have been written so far is returned. In the EINTR case the operation is
 * retried. Short writes are handled by issuing a subsequent write operation
 * for the remaining part.
 *
 * \return Negative on fatal errors, number of bytes written else.
 *
 * For blocking file descriptors, this function returns either the sum of all
 * buffer sizes or a negative error code which indicates the fatal error that
 * caused a write call to fail.
 *
 * For nonblocking file descriptors there is a third possibility: Any
 * non-negative return value less than the sum of the buffer sizes indicates
 * that a write operation returned EAGAIN/EWOULDBLOCK.
 *
 * \sa writev(2), \ref xwrite().
 */
int xwritev(int fd, struct iovec *iov, int iovcnt)
{
	size_t written = 0;
	int i;
	struct iovec saved_iov, *curiov;

	i = 0;
	curiov = iov;
	saved_iov = *curiov;
	while (i < iovcnt && curiov->iov_len > 0) {
		ssize_t ret = writev(fd, curiov, iovcnt - i);
		if (ret >= 0) {
			written += ret;
			while (ret > 0) {
				if (ret < curiov->iov_len) {
					curiov->iov_base += ret;
					curiov->iov_len -= ret;
					break;
				}
				ret -= curiov->iov_len;
				*curiov = saved_iov;
				i++;
				if (i >= iovcnt)
					return written;
				curiov++;
				saved_iov = *curiov;
			}
			continue;
		}
		if (errno == EINTR)
			/*
			 * The write() call was interrupted by a signal before
			 * any data was written. Try again.
			 */
			continue;
		if (errno == EAGAIN || errno == EWOULDBLOCK)
			/*
			 * We don't consider this an error. Note that POSIX
			 * allows either error to be returned, and does not
			 * require these constants to have the same value.
			 */
			return written;
		/* fatal error */
		return -ERRNO_TO_PARA_ERROR(errno);
	}
	return written;
}

/**
 * Write a buffer to a file descriptor, re-writing on short writes.
 *
 * \param fd The file descriptor.
 * \param buf The buffer to write.
 * \param len The number of bytes to write.
 *
 * This is a simple wrapper for \ref xwritev().
 *
 * \return The return value of the underlying call to \ref xwritev().
 */
int xwrite(int fd, const char *buf, size_t len)
{
	struct iovec iov = {.iov_base = (void *)buf, .iov_len = len};
	return xwritev(fd, &iov, 1);
}

/**
 * Write to a file descriptor, fail on short writes.
 *
 * \param fd The file descriptor.
 * \param buf The buffer to be written.
 * \param len The length of the buffer.
 *
 * For blocking file descriptors this function behaves identical to \ref
 * xwrite(). For non-blocking file descriptors it returns -E_SHORT_WRITE
 * (rather than a value less than len) if not all data could be written.
 *
 * \return Number of bytes written on success, negative error code else.
 */
int write_all(int fd, const char *buf, size_t len)
{
	int ret = xwrite(fd, buf, len);

	if (ret < 0)
		return ret;
	if (ret != len)
		return -E_SHORT_WRITE;
	return ret;
}

/**
 * A fprintf-like function for raw file descriptors.
 *
 * This function creates a string buffer according to the given format and
 * writes this buffer to a file descriptor.
 *
 * \param fd The file descriptor.
 * \param fmt A format string.
 *
 * The difference to fprintf(3) is that the first argument is a file
 * descriptor, not a FILE pointer. This function does not rely on stdio.
 *
 * \return The return value of the underlying call to \ref write_all().
 *
 * \sa fprintf(3), \ref xvasprintf().
 */
__printf_2_3 int write_va_buffer(int fd, const char *fmt, ...)
{
	char *msg;
	int ret;
	va_list ap;

	va_start(ap, fmt);
	ret = xvasprintf(&msg, fmt, ap);
	va_end(ap);
	ret = write_all(fd, msg, ret);
	free(msg);
	return ret;
}

/**
 * Read from a non-blocking file descriptor into multiple buffers.
 *
 * \param fd The file descriptor to read from.
 * \param iov Scatter/gather array used in readv().
 * \param iovcnt Number of elements in \a iov.
 * \param num_bytes Result pointer. Contains the number of bytes read from \a fd.
 *
 * This function tries to read up to sz bytes from fd, where sz is the sum of
 * the lengths of all vectors in iov. Like \ref xwrite(), EAGAIN and EINTR are
 * not considered error conditions. However, EOF is.
 *
 * \return Zero or a negative error code. If the underlying call to readv(2)
 * returned zero (indicating an end of file condition) or failed for some
 * reason other than EAGAIN or EINTR, a negative error code is returned.
 *
 * In any case, \a num_bytes contains the number of bytes that have been
 * successfully read from \a fd (zero if the first readv() call failed with
 * EAGAIN). Note that even if the function returns negative, some data might
 * have been read before the error occurred. In this case \a num_bytes is
 * positive.
 *
 * \sa \ref xwrite(), read(2), readv(2).
 */
int readv_nonblock(int fd, struct iovec *iov, int iovcnt, size_t *num_bytes)
{
	int ret, i, j;

	*num_bytes = 0;
	for (i = 0, j = 0; i < iovcnt;) {
		/* fix up the first iov */
		assert(j < iov[i].iov_len);
		iov[i].iov_base += j;
		iov[i].iov_len -= j;
		ret = readv(fd, iov + i, iovcnt - i);
		iov[i].iov_base -= j;
		iov[i].iov_len += j;

		if (ret == 0)
			return -E_EOF;
		if (ret < 0) {
			if (errno == EAGAIN || errno == EINTR)
				return 0;
			return -ERRNO_TO_PARA_ERROR(errno);
		}
		*num_bytes += ret;
		while (ret > 0) {
			if (ret < iov[i].iov_len - j) {
				j += ret;
				break;
			}
			ret -= iov[i].iov_len - j;
			j = 0;
			if (++i >= iovcnt)
				break;
		}
	}
	return 0;
}

/**
 * Read from a non-blocking file descriptor into a single buffer.
 *
 * \param fd The file descriptor to read from.
 * \param buf The buffer to read data to.
 * \param sz The size of \a buf.
 * \param num_bytes \see \ref readv_nonblock().
 *
 * This is a simple wrapper for readv_nonblock() which uses an iovec with a single
 * buffer.
 *
 * \return The return value of the underlying call to readv_nonblock().
 */
int read_nonblock(int fd, void *buf, size_t sz, size_t *num_bytes)
{
	struct iovec iov = {.iov_base = buf, .iov_len = sz};
	return readv_nonblock(fd, &iov, 1, num_bytes);
}

/**
 * Read a buffer and compare its contents to a string, ignoring case.
 *
 * \param fd The file descriptor to read from.
 * \param expectation The expected string to compare to.
 *
 * The given file descriptor is expected to be in non-blocking mode. The string
 * comparison is performed using strncasecmp(3).
 *
 * \return Zero if no data was available, positive if a buffer was read whose
 * contents compare as equal to the expected string, negative otherwise.
 * Possible errors: (a) not enough data was read, (b) the buffer contents
 * compared as non-equal, (c) a read error occurred. In the first two cases,
 * -E_READ_PATTERN is returned. In the read error case the (negative) return
 * value of the underlying call to \ref read_nonblock() is returned.
 */
int read_and_compare(int fd, const char *expectation)
{
	size_t n, len = strlen(expectation);
	char *buf = alloc(len + 1);
	int ret = read_nonblock(fd, buf, len, &n);

	if (ret < 0)
		goto out;
	buf[n] = '\0';
	ret = 0;
	if (n == 0)
		goto out;
	ret = -E_READ_PATTERN;
	if (n < len)
		goto out;
	if (strncasecmp(buf, expectation, len) != 0)
		goto out;
	ret = 1;
out:
	free(buf);
	return ret;
}

/**
 * Set a file descriptor to blocking mode.
 *
 * \param fd The file descriptor.
 *
 * \return Standard.
 */
__must_check int mark_fd_blocking(int fd)
{
	int flags = fcntl(fd, F_GETFL);
	if (flags < 0)
		return -ERRNO_TO_PARA_ERROR(errno);
	flags = fcntl(fd, F_SETFL, ((long)flags) & ~O_NONBLOCK);
	if (flags < 0)
		return -ERRNO_TO_PARA_ERROR(errno);
	return 1;
}

/**
 * Set a file descriptor to non-blocking mode.
 *
 * \param fd The file descriptor.
 *
 * \return Standard.
 */
__must_check int mark_fd_nonblocking(int fd)
{
	int flags = fcntl(fd, F_GETFL);
	if (flags < 0)
		return -ERRNO_TO_PARA_ERROR(errno);
	flags = fcntl(fd, F_SETFL, ((long)flags) | O_NONBLOCK);
	if (flags < 0)
		return -ERRNO_TO_PARA_ERROR(errno);
	return 1;
}

/**
 * Paraslash's wrapper for mmap.
 *
 * \param length Number of bytes to mmap.
 * \param prot Either PROT_NONE or the bitwise OR of one or more of
 * PROT_EXEC PROT_READ PROT_WRITE.
 * \param flags Exactly one of MAP_SHARED and MAP_PRIVATE.
 * \param fd The file to mmap from.
 * \param map Result pointer.
 *
 * \return Standard.
 *
 * \sa mmap(2).
 */
int para_mmap(size_t length, int prot, int flags, int fd, void *map)
{
	void **m = map;

	errno = EINVAL;
	if (!length)
		goto err;
	*m = mmap(NULL, length, prot, flags, fd, (off_t)0);
	if (*m != MAP_FAILED)
		return 1;
err:
	*m = NULL;
	return -ERRNO_TO_PARA_ERROR(errno);
}

/**
 * Wrapper for the open(2) system call.
 *
 * \param path The filename.
 * \param flags The usual open(2) flags.
 * \param mode Specifies the permissions to use.
 *
 * The mode parameter must be specified when O_CREAT is in the flags, and is
 * ignored otherwise.
 *
 * \return The file descriptor on success, negative on errors.
 *
 * \sa open(2).
 */
int para_open(const char *path, int flags, mode_t mode)
{
	int ret = open(path, flags, mode);

	if (ret >= 0)
		return ret;
	return -ERRNO_TO_PARA_ERROR(errno);
}

/**
 * Create a directory, don't fail if it already exists.
 *
 * \param path Name of the directory to create.
 *
 * This function passes the fixed mode value 0777 to mkdir(3) (which consults
 * the file creation mask and restricts this value).
 *
 * \return Zero if the path already existed as a directory or as a symbolic
 * link which leads to a directory, one if the path did not exist and the
 * directory has been created successfully, negative error code else.
 */
int para_mkdir(const char *path)
{
	/*
	 * We call opendir(3) rather than relying on stat(2) because this way
	 * we don't need extra code to get the symlink case right.
	 */
	DIR *dir = opendir(path);

	if (dir) {
		closedir(dir);
		return 0;
	}
	if (errno != ENOENT)
		return -ERRNO_TO_PARA_ERROR(errno);
	return mkdir(path, 0777) == 0? 1 : -ERRNO_TO_PARA_ERROR(errno);
}

/**
 * Open a file and map it into memory.
 *
 * \param path Name of the regular file to map.
 * \param open_mode Either \p O_RDONLY or \p O_RDWR.
 * \param map On success, the mapping is returned here.
 * \param size size of the mapping.
 * \param fd_ptr The file descriptor of the mapping.
 *
 * If \a fd_ptr is \p NULL, the file descriptor resulting from the underlying
 * open call is closed after mmap().  Otherwise the file is kept open and the
 * file descriptor is returned in \a fd_ptr.
 *
 * \return Standard.
 *
 * \sa para_open(), mmap(2).
 */
int mmap_full_file(const char *path, int open_mode, void **map,
		size_t *size, int *fd_ptr)
{
	int fd, ret, mmap_prot, mmap_flags;
	struct stat file_status;

	if (open_mode == O_RDONLY) {
		mmap_prot = PROT_READ;
		mmap_flags = MAP_PRIVATE;
	} else {
		mmap_prot = PROT_READ | PROT_WRITE;
		mmap_flags = MAP_SHARED;
	}
	ret = para_open(path, open_mode, 0);
	if (ret < 0)
		return ret;
	fd = ret;
	if (fstat(fd, &file_status) < 0) {
		ret = -ERRNO_TO_PARA_ERROR(errno);
		goto out;
	}
	*size = file_status.st_size;
	/*
	 * If the file is empty, *size is zero and mmap() would return EINVAL
	 * (Invalid argument). This error is common enough to spend an extra
	 * error code which explicitly states the problem.
	 */
	ret = -E_EMPTY;
	if (*size == 0)
		goto out;
	/*
	 * If fd refers to a directory, mmap() returns ENODEV (No such device),
	 * at least on Linux. "Is a directory" seems to be more to the point.
	 */
	ret = -ERRNO_TO_PARA_ERROR(EISDIR);
	if (S_ISDIR(file_status.st_mode))
		goto out;

	ret = para_mmap(*size, mmap_prot, mmap_flags, fd, map);
out:
	if (ret < 0 || !fd_ptr)
		close(fd);
	else
		*fd_ptr = fd;
	return ret;
}

/**
 * A wrapper for munmap(2).
 *
 * \param start The start address of the memory mapping.
 * \param length The size of the mapping.
 *
 * If NULL is passed as the start address, the length value is ignored and the
 * function does nothing.
 *
 * \return Zero if NULL was passed, one if the memory area was successfully
 * unmapped, a negative error code otherwise.
 *
 * \sa munmap(2), \ref mmap_full_file().
 */
int para_munmap(void *start, size_t length)
{
	if (!start)
		return 0;
	if (munmap(start, length) >= 0)
		return 1;
	return -ERRNO_TO_PARA_ERROR(errno);
}

/**
 * Simple wrapper for poll(2).
 *
 * It calls poll(2) and starts over if the call was interrupted by a signal.
 *
 * \param fds See poll(2).
 * \param nfds See poll(2).
 * \param timeout See poll(2).
 *
 * \return The return value of the underlying poll() call on success, the
 * negative paraslash error code on errors.
 *
 * All arguments are passed verbatim to poll(2).
 */
int xpoll(struct pollfd *fds, nfds_t nfds, int timeout)
{
	int ret;

	do
		ret = poll(fds, nfds, timeout);
	while (ret < 0 && errno == EINTR);
	return ret < 0? -ERRNO_TO_PARA_ERROR(errno) : ret;
}

/**
 * Check a file descriptor for readability.
 *
 * \param fd The file descriptor.
 *
 * \return positive if fd is ready for reading, zero if it isn't, negative if
 * an error occurred.
 *
 * \sa \ref write_ok().
 */
int read_ok(int fd)
{
	struct pollfd pfd = {.fd = fd, .events = POLLIN};
	int ret = xpoll(&pfd, 1, 0);
	return ret < 0? ret : pfd.revents & POLLIN;
}

/**
 * Check a file descriptor for writability.
 *
 * \param fd The file descriptor.
 *
 * \return positive if fd is ready for writing, zero if it isn't, negative if
 * an error occurred.
 *
 * \sa \ref read_ok().
 */
int write_ok(int fd)
{
	struct pollfd pfd = {.fd = fd, .events = POLLOUT};
	int ret = xpoll(&pfd, 1, 0);
	return ret < 0? ret : pfd.revents & POLLOUT;
}

/**
 * Ensure that file descriptors 0, 1, and 2 are valid.
 *
 * Common approach that opens /dev/null until it gets a file descriptor greater
 * than two.
 */
void valid_fd_012(void)
{
	while (1) {
		int fd = open("/dev/null", O_RDWR);
		if (fd < 0)
			exit(EXIT_FAILURE);
		if (fd > 2) {
			close(fd);
			break;
		}
	}
}
