/* Copyright (C) 2005 Andre Noll <maan@tuebingen.mpg.de>, see file COPYING. */

/** \file filter.c The stand-alone filter program. */

#include <lopsub.h>

#include "filter.lsg.h"
#include "filter_cmd.lsg.h"
#include "para.h"
#include "list.h"
#include "lsu.h"
#include "sched.h"
#include "buffer_tree.h"
#include "filter.h"
#include "string.h"
#include "stdin.h"
#include "stdout.h"
#include "error.h"
#include "fd.h"

/** Array of error strings. */
DEFINE_PARA_ERRLIST;
static struct lls_parse_result *lpr; /* command line options */

#define CMD_PTR (lls_cmd(0, filter_suite))
#define OPT_RESULT(_name) \
	(lls_opt_result(LSG_FILTER_PARA_FILTER_OPT_ ## _name, lpr))
#define OPT_GIVEN(_name) (lls_opt_given(OPT_RESULT(_name)))
#define OPT_UINT32_VAL(_name) (lls_uint32_val(0, OPT_RESULT(_name)))

/**
 * Dummy version which only contains NULL pointers.
 *
 * This is used by the amp filter which first tries to obtain the amplification
 * value from an element in this array.
 */
const char *stat_item_values[NUM_STAT_ITEMS] = {NULL};

/** The task that reads from stdin. */
static struct stdin_task stdin_task_struct;
/** pointer to the stdin task. */
static struct stdin_task *sit = &stdin_task_struct;

/** The task that writes converted data to stdout. */
static struct stdout_task stdout_task_struct;
/** Pointer to the stdout task. */
static struct stdout_task *sot = &stdout_task_struct;

static int loglevel;
INIT_STDERR_LOGGING(loglevel);

static void handle_help_flag(void)
{
	char *help;

	if (OPT_GIVEN(DETAILED_HELP))
		help = lls_long_help(CMD_PTR);
	else if (OPT_GIVEN(HELP))
		help = lls_short_help(CMD_PTR);
	else
		return;
	printf("%s\n", help);
	free(help);
	print_filter_helps(OPT_GIVEN(DETAILED_HELP));
	exit(EXIT_SUCCESS);
}

static int parse_config(void)
{
	int ret;

	version_handle_flag("filter", OPT_GIVEN(VERSION));
	handle_help_flag();
	ret = lsu_merge_config_file_options(NULL, "filter.conf",
		&lpr, CMD_PTR, filter_suite, 0 /* default flags */);
	if (ret < 0)
		return ret;
	loglevel = OPT_UINT32_VAL(LOGLEVEL);
	if (!OPT_GIVEN(FILTER)) {
		print_filter_list();
		exit(EXIT_SUCCESS);
	}
	return 1;
}

/**
 * The main function of para_filter.
 *
 * Para_filter reads data from stdin, converts it by using a chain
 * of filters (specified on the command line) and writes the resulting
 * data to stdout.
 *
 * \param argc Number of command line options.
 * \param argv Vector of arguments.
 *
 * \return \a EXIT_SUCCESS on success, EXIT_FAILURE on errors.
 */
int main(int argc, char *argv[])
{
	static struct sched s;
	int i, ret;
	const struct filter *f;
	struct btr_node *parent;
	struct filter_node **fns;
	struct lls_parse_result *filter_lpr; /* per-filter options */
	char *errctx;

	ret = lls(lls_parse(argc, argv, CMD_PTR, &lpr, &errctx));
	if (ret < 0)
		goto out;
	ret = parse_config();
	if (ret < 0)
		goto free_lpr;
	sit->btrn = btr_new_node(&(struct btr_node_description)
		EMBRACE(.name = "stdin"));
	stdin_task_register(sit, &s);

	fns = arr_alloc(OPT_GIVEN(FILTER), sizeof(*fns));
	for (i = 0, parent = sit->btrn; i < OPT_GIVEN(FILTER); i++) {
		const char *fa = lls_string_val(i, OPT_RESULT(FILTER));
		const char *name;
		struct filter_node *fn;
		struct task_info ti;

		fn = fns[i] = zalloc(sizeof(*fn));
		ret = filter_setup(fa, &fn->conf, &filter_lpr);
		if (ret < 0)
			goto teardown;
		fn->filter_num = ret;
		name = filter_name(fn->filter_num);
		fn->lpr = filter_lpr;
		PARA_DEBUG_LOG("filter #%d: %s\n", i, name);
		f = filter_get(fn->filter_num);
		fn->btrn = btr_new_node(&(struct btr_node_description)
			EMBRACE(.name = name, .parent = parent,
			.handler = f->execute, .context = fn));
		ti.name = name;
		ti.pre_monitor = f->pre_monitor;
		ti.post_monitor = f->post_monitor;
		ti.context = fn;
		if (f->open)
			f->open(fn);
		fn->task = task_register(&ti, &s);
		parent = fn->btrn;
	}
	sot->btrn = btr_new_node(&(struct btr_node_description)
		EMBRACE(.name = "stdout", .parent = parent));
	stdout_task_register(sot, &s);

	s.default_timeout = 1000;
	btr_log_tree(sit->btrn, LL_INFO);
	ret = schedule(&s);
	sched_shutdown(&s);
teardown:
	for (i--; i >= 0; i--) {
		struct filter_node *fn = fns[i];

		f = filter_get(fn->filter_num);
		if (f->close)
			f->close(fn);
		btr_remove_node(&fn->btrn);
		if (f->teardown)
			f->teardown(fn->lpr, fn->conf);
		free(fn);
	}
	free(fns);
	btr_remove_node(&sit->btrn);
	btr_remove_node(&sot->btrn);
free_lpr:
	lls_free_parse_result(lpr, CMD_PTR);
out:
	if (ret < 0) {
		if (errctx)
			PARA_ERROR_LOG("%s\n", errctx);
		free(errctx);
		PARA_ERROR_LOG("%s\n", para_strerror(-ret));
	}
	exit(ret < 0? EXIT_FAILURE : EXIT_SUCCESS);
}
