/* Copyright (C) 2007 Andre Noll <maan@tuebingen.mpg.de>, see file COPYING. */

/** \file mood.c Paraslash's mood handling functions. */

#include <osl.h>
#include <lopsub.h>

#include "para.h"
#include "error.h"
#include "string.h"
#include "afh.h"
#include "afs.h"
#include "list.h"

/*
 * Mood parser API. It's overkill to have an own header file for
 * these declarations as they are only needed in this .c file.
 */
struct mp_context;
int mp_init(const char *definition, int nbytes, struct mp_context **result,
		 char **errmsg);
bool mp_eval_row(const struct osl_row *aft_row, struct mp_context *ctx);
void mp_shutdown(struct mp_context *ctx);

/**
 * Contains statistical data of the currently admissible audio files.
 *
 * It is used to assign normalized score values to each admissible audio file.
 */
struct afs_statistics {
	/** Sum of num played over all admissible files. */
	int64_t num_played_sum;
	/** Sum of last played times over all admissible files. */
	int64_t last_played_sum;
	/** Quadratic deviation of num played count. */
	int64_t num_played_qd;
	/** Quadratic deviation of last played time. */
	int64_t last_played_qd;
	/** Correction factor for the num played score. */
	int64_t num_played_correction;
	/** Correction factor for the last played score. */
	int64_t last_played_correction;
	/** Common divisor of the correction factors. */
	int64_t normalization_divisor;
	/** Number of admissible files */
	unsigned num;
};

/**
 * Stores an instance of a loaded mood (parser and statistics).
 *
 * A structure of this type is allocated and initialized when a mood is loaded.
 */
struct mood_instance {
	/** NULL means that this is the "dummy" mood. */
	char *name;
	/** Bison's abstract syntax tree, used to determine admissibility. */
	struct mp_context *parser_context;
	/** To compute the score. */
	struct afs_statistics stats;
	/** NULL means to operate on the global score table. */
	struct osl_table *score_table;
};

/*
 * If current_mood is NULL then no mood is currently loaded. If
 * current_mood->name is NULL, the current mood is the dummy mood.
 *
 * The statistics are adjusted dynamically through this pointer as files are
 * added, removed or played.
 */
static struct mood_instance *current_mood;

/*
 * Find the position of the most-significant set bit.
 *
 * Copied and slightly adapted from the linux source tree, version 4.9.39
 * (2017-07).
 */
__a_const static uint32_t fls64(uint64_t v)
{
	int n = 63;
	const uint64_t ones = ~(uint64_t)0U;

	if ((v & (ones << 32)) == 0) {
		n -= 32;
		v <<= 32;
	}
	if ((v & (ones << (64 - 16))) == 0) {
		n -= 16;
		v <<= 16;
	}
	if ((v & (ones << (64 - 8))) == 0) {
		n -= 8;
		v <<= 8;
	}
	if ((v & (ones << (64 - 4))) == 0) {
		n -= 4;
		v <<= 4;
	}
	if ((v & (ones << (64 - 2))) == 0) {
		n -= 2;
		v <<= 2;
	}
	if ((v & (ones << (64 - 1))) == 0)
		n -= 1;
	return n;
}

/*
 * Compute the integer square root floor(sqrt(x)).
 *
 * Taken 2007 from the linux source tree.
 */
__a_const static uint64_t int_sqrt(uint64_t x)
{
	uint64_t op = x, res = 0, one = 1;

	one = one << (fls64(x) & ~one);
	while (one != 0) {
		if (op >= res + one) {
			op = op - (res + one);
			res = res + 2 * one;
		}
		res /= 2;
		one /= 4;
	}
	return res;
}

static void destroy_mood(struct mood_instance *m)
{
	if (!m)
		return;
	mp_shutdown(m->parser_context);
	if (m->score_table)
		score_close(m->score_table);
	free(m->name);
	free(m);
}

static struct mood_instance *alloc_new_mood(const char *name)
{
	struct mood_instance *m = zalloc(sizeof(*m));

	if (name)
		m->name = para_strdup(name);
	m->stats.normalization_divisor = 1;
	return m;
}

static int init_mood_parser(const char *mood_name, struct mood_instance **m,
		char **err)
{
	struct osl_object mood_def;
	int ret;

	if (!*mood_name) {
		if (err)
			*err = make_message("empty mood name\n");
		return -ERRNO_TO_PARA_ERROR(EINVAL);
	}
	ret = mood_get_def_by_name(mood_name, &mood_def);
	if (ret < 0) {
		if (err)
			*err = make_message("could not read mood definition\n");
		return ret;
	}
	*m = alloc_new_mood(mood_name);
	PARA_INFO_LOG("loading mood %s\n", mood_name);
	ret = mp_init(mood_def.data, mood_def.size, &(*m)->parser_context, err);
	osl_close_disk_object(&mood_def);
	if (ret < 0)
		destroy_mood(*m);
	return ret;
}

static int check_mood(struct osl_row *mood_row, void *data)
{
	struct afs_callback_arg *aca = data;
	char *mood_name, *errmsg;
	struct osl_object mood_def;
	struct mood_instance *m;
	int ret = mood_get_name_and_def_by_row(mood_row, &mood_name, &mood_def);

	if (ret < 0) {
		para_printf(&aca->pbout, "cannot read mood\n");
		return ret;
	}
	if (!*mood_name) /* ignore dummy row */
		goto out;
	para_printf(&aca->pbout, "checking mood %s\n", mood_name);
	m = alloc_new_mood("check");
	ret = mp_init(mood_def.data, mood_def.size, &m->parser_context,
		&errmsg);
	if (ret < 0) {
		para_printf(&aca->pbout, "%s: %s\n%s\n", mood_name, errmsg,
			para_strerror(-ret));
		free(errmsg);
	} else
		destroy_mood(m);
	ret = 1; /* don't fail the loop on invalid mood definitions */
out:
	osl_close_disk_object(&mood_def);
	return ret;
}

/**
 * Check all moods for syntax errors.
 *
 * \param aca Output goes to ->pbout, errors to ->fd on the error band.
 *
 * \return Negative on fatal errors. Inconsistent mood definitions are not
 * considered an error.
 */
int mood_check_callback(struct afs_callback_arg *aca)
{
	para_printf(&aca->pbout, "checking moods...\n");
	return osl(osl_rbtree_loop(moods_table, BLOBCOL_ID, aca, check_mood));
}

/*
 * The normalized num_played and last_played values are defined as
 *
 *	nn := -(np - mean_n) / sigma_n and nl := -(lp - mean_l) / sigma_l
 *
 *  For a (hypothetical) file with np = 0 and lp = now we thus have
 *
 *	nn =  mean_n / sigma_n =: hn > 0
 *	nl = -(now - mean_l) / sigma_l =: hl < 0
 *
 * We design the score function so that both contributions get the same
 * weight. Define the np and lp score of an arbitrary file as
 *
 *	sn := nn * -hl and sl := nl * hn
 *
 * Example:
 *	num_played mean/sigma: 87/14
 *	last_played mean/sigma: 45/32 days
 *
 *	We have hn = 87 / 14 = 6.21 and hl = -45 / 32 = -1.41. Multiplying
 *	nn of every file with the correction factor 1.41 and nl with
 *	6.21 makes the weight of the two contributions equal.
 *
 * The total score s := sn + sl has the representation
 *
 *	s = -cn * (np - mean_n) - cl * (lp - mean_l)
 *
 * with positive correction factors
 *
 *	cn = (now - mean_l) / (sqrt(ql) * sqrt(qn) / n)
 *	cl = mean_n / (sqrt(ql) * sqrt(qn) / n)
 *
 * where ql and qn are the quadratic deviations stored in the statistics
 * structure and n is the number of admissible files. To avoid integer
 * overflows and rounding errors we store the common divisor of the
 * correction factors separately.
 */
static long compute_score(struct afs_info *afsi,
		const struct afs_statistics *stats)
{
	int64_t mean_n, mean_l,score_n, score_l;

	assert(stats->normalization_divisor > 0);
	assert(stats->num > 0);
	mean_n = stats->num_played_sum / stats->num;
	mean_l = stats->last_played_sum / stats->num;

	score_n = -((int64_t)afsi->num_played - mean_n)
		* stats->num_played_correction
		/ stats->normalization_divisor;
	score_l = -((int64_t)afsi->last_played - mean_l)
		* stats->last_played_correction
		/ stats->normalization_divisor;
	return (score_n + score_l) / 2;
}

static int add_afs_statistics(const struct osl_row *row,
		struct afs_statistics *stats)
{
	uint64_t n, x, s, q;
	struct afs_info afsi;
	int ret;

	ret = get_afsi_of_row(row, &afsi);
	if (ret < 0)
		return ret;
	n = stats->num;
	x = afsi.last_played;
	s = stats->last_played_sum;
	if (n > 0) {
		q = (x > s / n)? x - s / n : s / n - x;
		stats->last_played_qd += q * q * n / (n + 1);
	}
	stats->last_played_sum += x;

	x = afsi.num_played;
	s = stats->num_played_sum;
	if (n > 0) {
		q = (x > s / n)? x - s / n : s / n - x;
		stats->num_played_qd += q * q * n / (n + 1);
	}
	stats->num_played_sum += x;
	stats->num++;
	return 1;
}

static int del_afs_statistics(const struct osl_row *row)
{
	struct afs_statistics *stats = &current_mood->stats;
	uint64_t n, s, q, a, new_s;
	struct afs_info afsi;
	int ret;
	ret = get_afsi_of_row(row, &afsi);
	if (ret < 0)
		return ret;
	n = stats->num;
	assert(n);
	if (n == 1) {
		memset(stats, 0, sizeof(*stats));
		stats->normalization_divisor = 1;
		return 1;
	}

	s = stats->last_played_sum;
	q = stats->last_played_qd;
	a = afsi.last_played;
	new_s = s - a;
	stats->last_played_sum = new_s;
	stats->last_played_qd = q + s * s / n - a * a
		- new_s * new_s / (n - 1);

	s = stats->num_played_sum;
	q = stats->num_played_qd;
	a = afsi.num_played;
	new_s = s - a;
	stats->num_played_sum = new_s;
	stats->num_played_qd = q + s * s / n - a * a
		- new_s * new_s / (n - 1);

	stats->num--;
	return 1;
}

/*
 * At mood load time we determine the set of admissible files for the given
 * mood where each file is identified by a pointer to a row of the audio file
 * table. In the first pass the pointers are added to a temporary array and
 * statistics are computed. When all admissible files have been processed in
 * this way, the score of each admissible file is computed and the (row, score)
 * pair is added to the score table. This has to be done in a second pass
 * since the score depends on the statistics. Finally, the array is freed.
 */
struct admissible_array {
	/** Files are admissible wrt. this mood. */
	struct mood_instance *m;
	/** The size of the array */
	unsigned size;
	/** Pointer to the array of admissible files. */
	struct osl_row **array;
};

/*
 * Check whether the given audio file is admissible. If it is, add it to array
 * of admissible files.
 */
static int add_if_admissible(struct osl_row *aft_row, void *data)
{
	struct admissible_array *aa = data;
	struct afs_statistics *stats = &aa->m->stats;

	if (!mp_eval_row(aft_row, aa->m->parser_context))
		return 0;
	if (stats->num >= aa->size) {
		aa->size *= 2;
		aa->size += 100;
		aa->array = arr_realloc(aa->array, aa->size,
			sizeof(struct osl_row *));
	}
	aa->array[stats->num] = aft_row;
	return add_afs_statistics(aft_row, stats);
}

/**
 * Compute the new quadratic deviation in case one element changes.
 *
 * \param n Number of elements.
 * \param old_qd The quadratic deviation before the change.
 * \param old_val The value that was replaced.
 * \param new_val The replacement value.
 * \param old_sum The sum of all elements before the update.
 *
 * \return The new quadratic deviation resulting from replacing old_val
 * by new_val.
 *
 * Given n real numbers a_1, ..., a_n, their sum S = a_1 + ... + a_n,
 * their quadratic deviation
 *
 * q = (a_1 - S/n)^2 + ... + (a_n - S/n)^2,
 *
 * and a real number b, the quadratic deviation q' of a_1,...a_{n-1}, b (ie.
 * the last number a_n was replaced by b) may be computed in O(1) time in terms
 * of n, q, a_n, b, and S as
 *
 * 	q' = q + d * s - (2 * S + d) * d / n
 *	   = q + d * (s - 2 * S / n - d /n),
 *
 * where d = b - a_n, and s = b + a_n.
 *
 * Example: n = 3, a_1 = 3, a_2 = 5, a_3 = 7, b = 10. Then S = 15, q = 8, d = 3,
 * s = 17, so
 *
 * 	q + d * s - (2 * S + d) * d / n = 8 + 51 - 33 = 26,
 *
 * which equals q' = (3 - 6)^2 + (5 - 6)^2 + (10 - 6)^2.
 *
 */
_static_inline_ int64_t update_quadratic_deviation(int64_t n, int64_t old_qd,
		int64_t old_val, int64_t new_val, int64_t old_sum)
{
	int64_t delta = new_val - old_val;
	int64_t sigma = new_val + old_val;
	return old_qd + delta * (sigma - 2 * old_sum / n - delta / n);
}

static void update_afs_statistics(struct afs_info *old_afsi,
		struct afs_info *new_afsi)
{
	struct afs_statistics *stats = &current_mood->stats;

	assert(stats->num > 0);
	stats->last_played_qd = update_quadratic_deviation(stats->num,
		stats->last_played_qd, old_afsi->last_played,
		new_afsi->last_played, stats->last_played_sum);
	stats->last_played_sum += new_afsi->last_played - old_afsi->last_played;

	stats->num_played_qd = update_quadratic_deviation(stats->num,
		stats->num_played_qd, old_afsi->num_played,
		new_afsi->num_played, stats->num_played_sum);
	stats->num_played_sum += new_afsi->num_played - old_afsi->num_played;
}

static int add_to_score_table(const struct osl_row *aft_row,
		struct mood_instance *m)
{
	long score;
	struct afs_info afsi;
	int ret = get_afsi_of_row(aft_row, &afsi);

	if (ret < 0)
		return ret;
	score = compute_score(&afsi, &m->stats);
	return score_add(aft_row, score, m->score_table);
}

static int delete_from_statistics_and_score_table(const struct osl_row *aft_row)
{
	int ret = del_afs_statistics(aft_row);
	if (ret < 0)
		return ret;
	return score_delete(aft_row);
}

/**
 * Delete an audio file from the score table and update mood statistics.
 *
 * \param aft_row Identifies the row to delete.
 *
 * \return Standard.
 *
 * \sa \ref score_delete().
 */
static int mood_delete_audio_file(const struct osl_row *aft_row)
{
	if (!row_belongs_to_score_table(aft_row))
		return 0;
	return delete_from_statistics_and_score_table(aft_row);
}

/**
 * Compute the new score of an audio file wrt. the current mood.
 *
 * \param aft_row Determines the audio file.
 * \param old_afsi The audio file selector info before updating.
 *
 * The \a old_afsi argument may be \p NULL which indicates that no changes to
 * the audio file info were made.
 *
 * \return Positive on success, negative on errors.
 */
static int mood_update_audio_file(const struct osl_row *aft_row,
		struct afs_info *old_afsi)
{
	long score, percent;
	int ret;
	bool is_admissible, was_admissible;
	struct afs_info afsi;

	if (!current_mood)
		return 1; /* nothing to do */
	was_admissible = row_belongs_to_score_table(aft_row);
	is_admissible = mp_eval_row(aft_row, current_mood->parser_context);
	if (!was_admissible && !is_admissible)
		return 1;
	if (was_admissible && !is_admissible)
		return delete_from_statistics_and_score_table(aft_row);
	if (!was_admissible && is_admissible) {
		ret = add_afs_statistics(aft_row, &current_mood->stats);
		if (ret < 0)
			return ret;
		return add_to_score_table(aft_row, current_mood);
	}
	/* update score */
	ret = get_afsi_of_row(aft_row, &afsi);
	if (ret < 0)
		return ret;
	if (old_afsi)
		update_afs_statistics(old_afsi, &afsi);
	score = compute_score(&afsi, &current_mood->stats);
	PARA_DEBUG_LOG("score: %li\n", score);
	percent = (score + 100) / 3;
	if (percent > 100)
		percent = 100;
	else if (percent < 0)
		percent = 0;
	PARA_DEBUG_LOG("moving to %li%%\n", percent);
	return score_update(aft_row, percent);
}

/* sse: seconds since epoch. */
static char *get_statistics(struct mood_instance *m, int64_t sse)
{
	unsigned n = m->stats.num;
	int mean_days, sigma_days;

	if (n == 0)
		return make_message("no admissible files\n");
	mean_days = (sse - m->stats.last_played_sum / n) / 3600 / 24;
	sigma_days = int_sqrt(m->stats.last_played_qd / n) / 3600 / 24;
	return make_message(
		"loaded mood %s (%u files)\n"
		"last_played mean/sigma: %d/%d days\n"
		"num_played mean/sigma: %" PRId64 "/%" PRIu64 "\n"
		"correction factor ratio: %.2lf\n"
	,
		m->name? m->name : "(dummy)",
		n,
		mean_days, sigma_days,
		m->stats.num_played_sum / n,
		int_sqrt(m->stats.num_played_qd / n),
		86400.0 * m->stats.last_played_correction /
			m->stats.num_played_correction
	);
}

/**
 * Free all resources of a mood instance.
 *
 * \param m As obtained by \ref mood_load(). If NULL, unload the current mood.
 *
 * It's OK to call this with m == NULL even if no current mood is loaded.
 */
void mood_unload(struct mood_instance *m)
{
	if (m)
		return destroy_mood(m);
	destroy_mood(current_mood);
	current_mood = NULL;
}

static void compute_correction_factors(int64_t sse, struct afs_statistics *s)
{
	if (s->num > 0) {
		s->normalization_divisor = int_sqrt(s->last_played_qd)
			* int_sqrt(s->num_played_qd) / s->num / 100;
		s->num_played_correction = sse - s->last_played_sum / s->num;
		s->last_played_correction = s->num_played_sum / s->num;
	}
	if (s->num_played_correction == 0)
		s->num_played_correction = 1;
	if (s->normalization_divisor == 0)
		s->normalization_divisor = 1;
	if (s->last_played_correction == 0)
		s->last_played_correction = 1;
}

/**
 * Populate a score table with admissible files for the given mood.
 *
 * This consults the mood table to initialize the mood parser with the mood
 * expression stored in the blob object which corresponds to the given name. A
 * score table is allocated and populated with references to those entries of
 * the audio file table which evaluate as admissible with respect to the mood
 * expression. For each audio file a score value is computed and stored along
 * with the file reference.
 *
 * \param mood_name The name of the mood to load.
 * \param result Opaque, refers to the mood parser and the score table.
 * \param msg Error message or mood info is returned here.
 *
 * If the mood name is NULL, the dummy mood is loaded. This mood regards every
 * audio file as admissible.
 *
 * A NULL result pointer instructs the function to operate on the current mood.
 * That is, on the mood instance which is used by the server to select the next
 * audio file for streaming. In this mode of operation, the mood which was
 * active before the call, if any, is unloaded on success.
 *
 * If result is not NULL, the current mood is unaffected and *result points to
 * an initialized mood instance on success. The caller can pass this reference
 * to \ref mood_loop() to iterate over the admissible files, and should call
 * \ref mood_unload() to free the mood instance afterwards.
 *
 * If the message pointer is not NULL, a suitable message is returned there in
 * all cases. The caller must free this string.
 *
 * \return Standard. On errors, the current mood remains unaffected even if
 * result is NULL. It is not considered an error if no files are admissible.
 *
 * \sa \ref mp_eval_row().
 */
int mood_load(const char *mood_name, struct mood_instance **result, char **msg)
{
	int i, ret;
	struct admissible_array aa = {.size = 0};
	/*
	 * We can not use the "now" pointer from sched.c here because we are
	 * called before schedule(), which initializes "now".
	 */
	struct timeval rnow;

	if (mood_name) {
		ret = init_mood_parser(mood_name, &aa.m, msg);
		if (ret < 0)
			return ret;
	} else /* load dummy mood */
		aa.m = alloc_new_mood(NULL);
	PARA_NOTICE_LOG("computing statistics of admissible files\n");
	ret = audio_file_loop(&aa, add_if_admissible);
	if (ret < 0) {
		if (msg) /* false if we are called via the event handler */
			*msg = make_message("audio file loop failed\n");
		goto out;
	}
	clock_get_realtime(&rnow);
	compute_correction_factors(rnow.tv_sec, &aa.m->stats);
	if (result)
		score_open(&aa.m->score_table);
	for (i = 0; i < aa.m->stats.num; i++) {
		ret = add_to_score_table(aa.array[i], aa.m);
		if (ret < 0) {
			if (msg)
				*msg = make_message(
					"could not add row to score table\n");
			goto out;
		}
	}
	/* success */
	if (msg)
		*msg = get_statistics(aa.m, rnow.tv_sec);
	if (result)
		*result = aa.m;
	else {
		mood_unload(NULL);
		current_mood = aa.m;
	}
	ret = 1;
out:
	free(aa.array);
	if (ret <= 0) /* error, or no admissible files */
		destroy_mood(aa.m);
	return ret;
}

/**
 * Iterate over the admissible files of a mood instance.
 *
 * This wrapper around \ref score_loop() is the mood counterpart of \ref
 * playlist_loop().
 *
 * \param m Determines the score table to iterate. Must not be NULL.
 * \param func See \ref score_loop().
 * \param data See \ref score_loop().
 *
 * \return See \ref score_loop(), \ref playlist_loop().
 */
int mood_loop(struct mood_instance *m, osl_rbtree_loop_func *func, void *data)
{
	return score_loop(func, m->score_table, data);
}

/*
 * Empty the score table and start over.
 *
 * This function is called on events which render the current set of admissible
 * files invalid, for example if an attribute is removed from the attribute
 * table.
 */
static int reload_current_mood(void)
{
	int ret;
	char *mood_name = NULL;

	assert(current_mood);
	score_clear();
	PARA_NOTICE_LOG("reloading %s\n", current_mood->name?
		current_mood->name : "(dummy)");
	if (current_mood->name)
		mood_name = para_strdup(current_mood->name);
	mood_unload(NULL);
	ret = mood_load(mood_name, NULL, NULL);
	free(mood_name);
	return ret;
}

/**
 * Notification callback for the moods table.
 *
 * \param event Type of the event just occurred.
 * \param pb Unused.
 * \param data Its type depends on the event.
 *
 * This function updates the score table according to the event that has
 * occurred. Two actions are possible: (a) reload the current mood, or (b)
 * add/remove/update the row of the score table which corresponds to the audio
 * file that has been modified or whose afs info has been changed. It depends
 * on the type of the event which action (if any) is performed.
 *
 * The callbacks of command handlers such as com_add() or com_touch() which
 * modify the audio file table call this function. The virtual streaming system
 * also calls this after it has updated the afs info of the file it is about to
 * stream (the one with the highest score). If the file stays admissible, its
 * score is recomputed so that a different file is picked next time.
 *
 * \return Standard.
 */
int moods_event_handler(enum afs_events event, __a_unused struct para_buffer *pb,
		void *data)
{
	if (!current_mood)
		return 0;
	switch (event) {
	/*
	 * The three blob events might change the set of admissible files,
	 * so we must reload the score list.
	 */
	case BLOB_RENAME:
	case BLOB_REMOVE:
	case BLOB_ADD:
		if (data == moods_table || data == playlists_table)
			return 1; /* no reload necessary for these */
		return reload_current_mood();
	/* these also require reload of the score table */
	case ATTRIBUTE_ADD:
	case ATTRIBUTE_REMOVE:
	case ATTRIBUTE_RENAME:
		return reload_current_mood();
	/* changes to the aft only require to re-examine the audio file */
	case VSS_NEW_AUDIO_FILE:
	case AFSI_CHANGE: {
		struct afsi_change_event_data *aced = data;
		return mood_update_audio_file(aced->aft_row, aced->old_afsi);
		}
	case AFHI_CHANGE:
	case AUDIO_FILE_RENAME:
	case AUDIO_FILE_ADD:
		return mood_update_audio_file(data, NULL);
	case AUDIO_FILE_REMOVE:
		return mood_delete_audio_file(data);
	default:
		return 1;
	}
}
