/* Copyright (C) 1997 Andre Noll <maan@tuebingen.mpg.de>, see file COPYING. */

/** \file server.c Paraslash's main server. */

#include <netinet/in.h>
#include <sys/socket.h>
#include <signal.h>
#include <osl.h>
#include <sys/types.h>
#include <arpa/inet.h>
#include <sys/un.h>
#include <netdb.h>
#include <lopsub.h>

#include "server.lsg.h"
#include "para.h"
#include "error.h"
#include "lsu.h"
#include "crypt.h"
#include "afh.h"
#include "string.h"
#include "afs.h"
#include "net.h"
#include "server.h"
#include "list.h"
#include "sched.h"
#include "send.h"
#include "vss.h"
#include "config.h"
#include "close_on_fork.h"
#include "daemon.h"
#include "ipc.h"
#include "fd.h"
#include "signal.h"
#include "user_list.h"
#include "color.h"

/** Array of error strings. */
DEFINE_PARA_ERRLIST;

__printf_2_3 void (*para_log)(int, const char*, ...) = daemon_log;

/** Shut down non-authorized connections after that many seconds. */
#define ALARM_TIMEOUT 10

/**
 * Pointer to shared memory area for communication between para_server
 * and its children. Exported to vss.c, command.c and to afs.
 */
struct misc_meta_data *mmd;

/**
 * The active value for all config options of para_server.
 *
 * It is computed by merging the parse result of the command line options with
 * the parse result of the config file.
 */
struct lls_parse_result *server_lpr = NULL;

/* Command line options (no config file options). Used in handle_sighup(). */
static struct lls_parse_result *cmdline_lpr;

/**
 * A random number used to "authenticate" the afs connection.
 *
 * para_server picks this number by random before it forks the afs process. The
 * command handlers know this number as well and write it to the afs socket,
 * together with the id of the shared memory area which contains the payload of
 * the afs command. A local process has to know this number to abuse the afs
 * service provided by the local socket.
 */
uint32_t afs_socket_cookie;

/** The mutex protecting the shared memory area containing the mmd struct. */
int mmd_mutex;

/* Serializes log output. */
static int log_mutex;

static struct sched sched;
static struct signal_task *signal_task;

/** The process id of the audio file selector process. */
pid_t afs_pid = 0;

/* The main server process (parent of afs and the command handlers). */
static pid_t server_pid;

/**
 * Tell whether the executing process is a command handler.
 *
 * Cleanup on exit must be performed differently for command handlers.
 *
 * \return True if the pid of the executing process is neither the server pid
 * nor the afs pid.
 */
bool process_is_command_handler(void)
{
	pid_t pid = getpid();

	return pid != afs_pid && pid != server_pid;
}

/** The task responsible for server command handling. */
struct server_command_task {
	unsigned num_listen_fds; /* only one by default */
	/** TCP socket(s) on which para_server listens for connections. */
	int *listen_fds;
	/* File descriptor for the accepted socket. */
	int child_fd;
	/** Copied from para_server's main function. */
	int argc;
	/** Argument vector passed to para_server's main function. */
	char **argv;
	/** The command task structure for scheduling. */
	struct task *task;
};

static void pre_log_hook(void)
{
	mutex_lock(log_mutex);
}

static void post_log_hook(void)
{
	mutex_unlock(log_mutex);
}

/* Setup shared memory area and init mutexes */
static void init_ipc_or_die(void)
{
	void *shm;
	int shmid, ret = shm_new(sizeof(struct misc_meta_data));

	if (ret < 0)
		goto err_out;
	shmid = ret;
	ret = shm_attach(shmid, ATTACH_RW, &shm);
	shm_destroy(shmid);
	if (ret < 0)
		goto err_out;
	mmd = shm;

	ret = mutex_new();
	if (ret < 0)
		goto err_out;
	mmd_mutex = ret;
	ret = mutex_new();
	if (ret < 0)
		goto destroy_mmd_mutex;
	log_mutex = ret;

	mmd->num_played = 0;
	mmd->num_commands = 0;
	mmd->events = 0;
	mmd->num_connects = 0;
	mmd->active_connections = 0;
	mmd->vss_status_flags = VSS_NEXT;
	mmd->new_vss_status_flags = VSS_NEXT;
	mmd->loglevel = OPT_UINT32_VAL(LOGLEVEL);
	return;
destroy_mmd_mutex:
	mutex_destroy(mmd_mutex);
err_out:
	PARA_EMERG_LOG("%s\n", para_strerror(-ret));
	exit(EXIT_FAILURE);
}

/** Get a reference to the supercommand of para_server. */
#define CMD_PTR (lls_cmd(0, server_suite))

/**
 * (Re-)read the server configuration files.
 *
 * \param reload Whether config file overrides command line.
 *
 * This function also re-opens the logfile and the user list. On SIGHUP it is
 * called from both server and afs context.
 */
void parse_config_or_die(bool reload)
{
	int ret;
	unsigned flags = MCF_DONT_FREE;

	if (server_lpr != cmdline_lpr)
		lls_free_parse_result(server_lpr, CMD_PTR);
	server_lpr = cmdline_lpr;
	if (reload)
		flags |= MCF_OVERRIDE;
	ret = lsu_merge_config_file_options(OPT_STRING_VAL(CONFIG_FILE),
		"server.conf", &server_lpr, CMD_PTR, server_suite, flags);
	if (ret < 0) {
		PARA_EMERG_LOG("failed to parse config file: %s\n",
			para_strerror(-ret));
		exit(EXIT_FAILURE);
	}
	daemon_set_loglevel(OPT_UINT32_VAL(LOGLEVEL));
	if (OPT_GIVEN(LOGFILE)) {
		daemon_set_logfile(OPT_STRING_VAL(LOGFILE));
		daemon_open_log_or_die();
	}
	if (daemon_init_colors_or_die(OPT_UINT32_VAL(COLOR), COLOR_AUTO,
			COLOR_NO, OPT_GIVEN(LOGFILE))) {
		int i;
		for (i = 0; i < OPT_GIVEN(LOG_COLOR); i++)
			daemon_set_log_color_or_die(lls_string_val(i,
				OPT_RESULT(LOG_COLOR)));
	}
	daemon_set_flag(DF_LOG_PID);
	daemon_set_flag(DF_LOG_LL);
	daemon_set_flag(DF_LOG_TIME);
	if (OPT_GIVEN(LOG_TIMING))
		daemon_set_flag(DF_LOG_TIMING);
	daemon_set_priority(OPT_UINT32_VAL(PRIORITY));
	if (!reload || getpid() != afs_pid) {
		char *user_list_file;
		if (OPT_GIVEN(USER_LIST))
			user_list_file = para_strdup(OPT_STRING_VAL(USER_LIST));
		else {
			char *home = para_homedir();
			user_list_file = make_message("%s/.paraslash/server.users", home);
			free(home);
		}
		user_list_init(user_list_file);
		free(user_list_file);
	}
	return;
}

/*
 * called when server gets SIGHUP or when client invokes hup command.
 */
static void handle_sighup(void)
{

	PARA_NOTICE_LOG("SIGHUP\n");
	parse_config_or_die(true);
	if (afs_pid != 0)
		kill(afs_pid, SIGHUP);
}

static int signal_post_monitor(struct sched *s, __a_unused void *context)
{
	int ret, signum;

	ret = task_get_notification(signal_task->task);
	if (ret < 0)
		return ret;
	signum = para_next_signal();
	switch (signum) {
	case 0:
		return 0;
	case SIGHUP:
		handle_sighup();
		break;
	case SIGCHLD:
		for (;;) {
			pid_t pid;
			ret = para_reap_child(&pid);
			if (ret <= 0)
				break;
			if (pid != afs_pid)
				continue;
			PARA_EMERG_LOG("fatal: afs died\n");
			goto genocide;
		}
		break;
	/* die on sigint/sigterm. Kill all children too. */
	case SIGINT:
	case SIGTERM:
		PARA_EMERG_LOG("terminating on signal %d\n", signum);
genocide:
		kill(0, SIGTERM);
		/*
		 * We must wait for all of our children to die. For the afs
		 * process or a command handler might want to use the
		 * shared memory area and the mmd mutex.  If we destroy this
		 * mutex too early and afs tries to lock the shared memory
		 * area, the call to mutex_lock() will fail and terminate the
		 * afs process. This leads to dirty osl tables.
		 */
		PARA_INFO_LOG("waiting for child processes to die\n");
		mutex_unlock(mmd_mutex);
		while (wait(NULL) != -1 || errno != ECHILD)
			; /* still at least one child alive */
		mutex_lock(mmd_mutex);
		free(mmd->afd.afhi.chunk_table);
		task_notify_all(s, E_DEADLY_SIGNAL);
		return -E_DEADLY_SIGNAL;
	}
	return 0;
}

static void init_signal_task(void)
{
	signal_task = signal_init_or_die();
	para_install_sighandler(SIGINT);
	para_install_sighandler(SIGTERM);
	para_install_sighandler(SIGHUP);
	para_install_sighandler(SIGCHLD);
	para_sigaction(SIGPIPE, SIG_IGN);
	add_close_on_fork_list(signal_task->fd);
	signal_task->task = task_register(&(struct task_info) {
		.name = "signal",
		.pre_monitor = signal_pre_monitor,
		.post_monitor = signal_post_monitor,
		.context = signal_task,

	}, &sched);
}

static void command_pre_monitor(struct sched *s, void *context)
{
	unsigned n;
	struct server_command_task *sct = context;

	for (n = 0; n < sct->num_listen_fds; n++)
		sched_monitor_readfd(sct->listen_fds[n], s);
}

static int command_task_accept(unsigned listen_idx, struct sched *s,
		struct server_command_task *sct)
{
	int new_fd, ret, i;
	char *peer_name;
	pid_t child_pid;
	uint32_t *chunk_table;

	ret = para_accept(sct->listen_fds[listen_idx], NULL, 0, &new_fd);
	if (ret <= 0)
		goto out;
	mmd->num_connects++;
	mmd->active_connections++;
	/*
	 * The chunk table is a pointer located in the mmd struct that points
	 * to dynamically allocated memory, i.e. it must be freed by the parent
	 * and the child. However, as the mmd struct is in a shared memory
	 * area, there's no guarantee that after the fork this pointer is still
	 * valid in child context. As it is not used in the child anyway, we
	 * save it to a local variable before the fork and free the memory via
	 * that copy in the child directly after the fork.
	 */
	chunk_table = mmd->afd.afhi.chunk_table;
	child_pid = fork();
	if (child_pid < 0) {
		ret = -ERRNO_TO_PARA_ERROR(errno);
		goto out;
	}
	if (child_pid) {
		/* avoid problems with non-fork-safe PRNGs */
		unsigned char buf[16];
		get_random_bytes_or_die(buf, sizeof(buf));
		close(new_fd);
		/* parent keeps accepting connections */
		return 0;
	}
	peer_name = remote_name(new_fd);
	PARA_INFO_LOG("accepted connection from %s\n", peer_name);
	/* mmd might already have changed at this point */
	free(chunk_table);
	sct->child_fd = new_fd;
	/*
	 * put info on who we are serving into argv[0] to make
	 * client ip visible in top/ps
	 */
	for (i = sct->argc - 1; i >= 0; i--)
		memset(sct->argv[i], 0, strlen(sct->argv[i]));
	i = sct->argc - 1 - lls_num_inputs(cmdline_lpr);
	sprintf(sct->argv[i], "para_server (serving %s)", peer_name);
	/* ask other tasks to terminate */
	task_notify_all(s, E_CHILD_CONTEXT);
	/*
	 * After we return, the scheduler calls server_select() with a minimal
	 * timeout value, because the remaining tasks have a notification
	 * pending. Next it calls the ->post_monitor method of these tasks,
	 * which will return negative in view of the notification. This causes
	 * schedule() to return as there are no more runnable tasks.
	 *
	 * Note that semaphores are not inherited across a fork(), so we don't
	 * hold the lock at this point. Since server_poll() drops the lock
	 * prior to calling poll(), we need to acquire it here.
	 */
	mutex_lock(mmd_mutex);
	return -E_CHILD_CONTEXT;
out:
	if (ret < 0)
		PARA_CRIT_LOG("%s\n", para_strerror(-ret));
	return 0;
}

static int command_post_monitor(struct sched *s, void *context)
{
	struct server_command_task *sct = context;
	unsigned n;
	int ret;

	ret = task_get_notification(sct->task);
	if (ret < 0)
		goto fail;
	for (n = 0; n < sct->num_listen_fds; n++) {
		ret = command_task_accept(n, s, sct);
		if (ret < 0)
			goto fail;
	}
	return 0;
fail:
	free(sct->listen_fds);
	return ret;
}

static void init_server_command_task(struct server_command_task *sct,
		int argc, char **argv)
{
	int ret;
	unsigned n;
	uint32_t port = OPT_UINT32_VAL(PORT);

	PARA_NOTICE_LOG("initializing tcp command socket\n");
	sct->child_fd = -1;
	sct->argc = argc;
	sct->argv = argv;
	if (!OPT_GIVEN(LISTEN_ADDRESS)) {
		sct->num_listen_fds = 1;
		sct->listen_fds = alloc(sizeof(int));
		ret = para_listen_simple(IPPROTO_TCP, port);
		if (ret < 0)
			goto err;
		sct->listen_fds[0] = ret;
	} else {
		sct->num_listen_fds = OPT_GIVEN(LISTEN_ADDRESS);
		sct->listen_fds = alloc(sct->num_listen_fds * sizeof(int));
		for (n = 0; n < OPT_GIVEN(LISTEN_ADDRESS); n++) {
			const char *arg;
			arg = lls_string_val(n, OPT_RESULT(LISTEN_ADDRESS));
			ret = para_listen(IPPROTO_TCP, arg, port);
			if (ret < 0)
				goto err;
			sct->listen_fds[n] = ret;
		}
	}
	for (n = 0; n < sct->num_listen_fds; n++) {
		ret = mark_fd_nonblocking(sct->listen_fds[n]);
		if (ret < 0)
			goto err;
		/* child doesn't need the listener */
		add_close_on_fork_list(sct->listen_fds[n]);
	}

	sct->task = task_register(&(struct task_info) {
		.name = "server command",
		.pre_monitor = command_pre_monitor,
		.post_monitor = command_post_monitor,
		.context = sct,
	}, &sched);
	/*
	 * Detect whether the abstract Unix domain socket space is supported,
	 * but do not create the socket. We check this once in server context
	 * so that the command handlers inherit this bit of information and
	 * don't need to check again.
	 */
	create_local_socket(NULL);
	return;
err:
	PARA_EMERG_LOG("%s\n", para_strerror(-ret));
	exit(EXIT_FAILURE);
}

static int init_afs(int argc, char **argv)
{
	int ret, afs_server_socket[2];
	char c;

	ret = socketpair(PF_UNIX, SOCK_STREAM, 0, afs_server_socket);
	if (ret < 0)
		exit(EXIT_FAILURE);
	get_random_bytes_or_die((unsigned char *)&afs_socket_cookie,
		sizeof(afs_socket_cookie));
	afs_pid = fork();
	if (afs_pid < 0)
		exit(EXIT_FAILURE);
	if (afs_pid == 0) { /* child (afs) */
		int i;

		afs_pid = getpid();
		crypt_shutdown();
		user_list_deplete();
		for (i = argc - 1; i >= 0; i--)
			memset(argv[i], 0, strlen(argv[i]));
		i = argc - lls_num_inputs(cmdline_lpr) - 1;
		sprintf(argv[i], "para_server (afs)");
		close(afs_server_socket[0]);
		afs_init(afs_server_socket[1]);
	}
	close(afs_server_socket[1]);
	if (read(afs_server_socket[0], &c, 1) <= 0) {
		PARA_EMERG_LOG("early afs exit\n");
		exit(EXIT_FAILURE);
	}
	ret = mark_fd_nonblocking(afs_server_socket[0]);
	if (ret < 0)
		exit(EXIT_FAILURE);
	add_close_on_fork_list(afs_server_socket[0]);
	PARA_INFO_LOG("afs_socket: %d, afs_socket_cookie: %u\n",
		afs_server_socket[0], (unsigned) afs_socket_cookie);
	return afs_server_socket[0];
}

static void handle_help_flags(void)
{
	char *help;
	bool d = OPT_GIVEN(DETAILED_HELP);

	if (d)
		help = lls_long_help(CMD_PTR);
	else if (OPT_GIVEN(HELP))
		help = lls_short_help(CMD_PTR);
	else
		return;
	printf("%s\n", help);
	free(help);
	exit(EXIT_SUCCESS);
}

static void server_init(int argc, char **argv, struct server_command_task *sct)
{
	int ret, afs_socket, daemon_pipe = -1;
	char *errctx;

	valid_fd_012();
	/* parse command line options */
	ret = lls(lls_parse(argc, argv, CMD_PTR, &cmdline_lpr, &errctx));
	if (ret < 0)
		goto fail;
	server_lpr = cmdline_lpr;
	daemon_set_loglevel(OPT_UINT32_VAL(LOGLEVEL));
	daemon_drop_privileges_or_die(OPT_STRING_VAL(USER),
		OPT_STRING_VAL(GROUP));
	version_handle_flag("server", OPT_GIVEN(VERSION));
	handle_help_flags();
	parse_config_or_die(false);
	/* become daemon */
	if (OPT_GIVEN(DAEMON))
		daemon_pipe = daemonize(true /* parent waits for SIGTERM */);
	server_pid = getpid();
	crypt_init();
	daemon_log_welcome("server");
	init_ipc_or_die(); /* init mmd struct, mmd and log mutex */
	daemon_set_start_time();
	daemon_set_hooks(pre_log_hook, post_log_hook);
	/*
	 * Although afs uses its own signal handling we must ignore SIGUSR1
	 * _before_ the afs child process gets born by init_afs() below.  It's
	 * racy to do this in the child because the parent might send SIGUSR1
	 * before the child gets a chance to ignore this signal.
	 *
	 * We also have to block SIGCHLD before the afs process is created
	 * because otherwise para_server does not notice if afs dies before the
	 * SIGCHLD handler has been installed for the parent process by
	 * init_signal_task() below.
	 */
	para_sigaction(SIGUSR1, SIG_IGN);
	para_block_signal(SIGCHLD);
	PARA_NOTICE_LOG("initializing the audio file selector\n");
	afs_socket = init_afs(argc, argv);
	init_signal_task();
	para_unblock_signal(SIGCHLD);
	PARA_NOTICE_LOG("initializing virtual streaming system\n");
	vss_init(afs_socket, &sched);
	init_server_command_task(sct, argc, argv);
	if (daemon_pipe >= 0) {
		if (write(daemon_pipe, "\0", 1) < 0) {
			PARA_EMERG_LOG("daemon_pipe: %s", strerror(errno));
			exit(EXIT_FAILURE);
		}
		close(daemon_pipe);
	}
	PARA_NOTICE_LOG("server init complete\n");
	return;
fail:
	assert(ret < 0);
	if (errctx)
		PARA_ERROR_LOG("%s\n", errctx);
	PARA_EMERG_LOG("%s\n", para_strerror(-ret));
	exit(EXIT_FAILURE);
}

static void status_refresh(void)
{
	static int prev_uptime = -1, prev_events = -1;
	int uptime = daemon_get_uptime(now);

	if (prev_events != mmd->events)
		goto out;
	if (mmd->new_vss_status_flags != mmd->vss_status_flags)
		goto out_inc_events;
	if (uptime / 60 != prev_uptime / 60)
		goto out_inc_events;
	return;
out_inc_events:
	mmd->events++;
out:
	prev_uptime = uptime;
	prev_events = mmd->events;
	mmd->vss_status_flags = mmd->new_vss_status_flags;
	PARA_DEBUG_LOG("%u events, forcing status update\n", mmd->events);
	killpg(0, SIGUSR1);
}

static int server_poll(struct pollfd *fds, nfds_t nfds, int timeout)
{
	int ret;

	daemon_set_loglevel(mmd->loglevel);
	status_refresh();
	mutex_unlock(mmd_mutex);
	ret = xpoll(fds, nfds, timeout);
	mutex_lock(mmd_mutex);
	return ret;
}

/**
 * Deallocate all lopsub parse results.
 *
 * The server allocates a parse result for command line options and optionally
 * a second parse result for the effective configuration, defined by merging
 * the command line options with the options stored in the configuration file.
 * This function frees both structures.
 */
void free_lpr(void)
{
	lls_free_parse_result(server_lpr, CMD_PTR);
	if (server_lpr != cmdline_lpr)
		lls_free_parse_result(cmdline_lpr, CMD_PTR);
}

/**
 * The main function of para_server.
 *
 * \param argc Usual argument count.
 * \param argv Usual argument vector.
 *
 * \return EXIT_SUCCESS or EXIT_FAILURE.
 */
int main(int argc, char *argv[])
{
	int ret;
	struct server_command_task server_command_task_struct,
		*sct = &server_command_task_struct;

	sched.default_timeout = 1000;
	sched.poll_function = server_poll;

	server_init(argc, argv, sct);
	mutex_lock(mmd_mutex);
	ret = schedule(&sched);
	/*
	 * We hold the mmd lock: it was re-acquired in server_poll()
	 * after the poll(2) call.
	 */
	mutex_unlock(mmd_mutex);
	sched_shutdown(&sched);
	crypt_shutdown();
	signal_shutdown(signal_task);
	if (!process_is_command_handler()) { /* parent (server) */
		mutex_destroy(mmd_mutex);
		daemon_set_hooks(NULL, NULL); /* only one process remaining */
		mutex_destroy(log_mutex);
		deplete_close_on_fork_list();
		if (ret < 0)
			PARA_EMERG_LOG("%s\n", para_strerror(-ret));
		vss_shutdown();
	} else {
		vss_shutdown();
		alarm(ALARM_TIMEOUT);
		close_listed_fds();
		ret = handle_connect(sct->child_fd);
	}
	shm_detach(mmd);
	user_list_deplete();
	free_lpr();
	exit(ret < 0? EXIT_FAILURE : EXIT_SUCCESS);
}
