/* Copyright (C) 2010 Andre Noll <maan@tuebingen.mpg.de>, see file COPYING. */

/* This file is based on speexdec.c, by Jean-Marc Valin, see below. */

/* Copyright (C) 2002-2006 Jean-Marc Valin
   File: speexdec.c

   Redistribution and use in source and binary forms, with or without
   modification, are permitted provided that the following conditions
   are met:
   
   - Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.
   
   - Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the distribution.
   
   - Neither the name of the Xiph.org Foundation nor the names of its
   contributors may be used to endorse or promote products derived from
   this software without specific prior written permission.
   
   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
   ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
   LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
   A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE FOUNDATION OR
   CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
   EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
   PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
   PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
   LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
   NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
   SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/
/** \file spx_afh.c Audio format handler for ogg/speex files. */

#include <ogg/ogg.h>
#include <speex/speex.h>
#include <speex/speex_header.h>
#include <speex/speex_stereo.h>

#include "para.h"
#include "afh.h"
#include "error.h"
#include "portable_io.h"
#include "string.h"
#include "spx.h"
#include "ogg_afh_common.h"

struct private_spx_data {
	struct spx_header_info shi;
};

static bool copy_if_tag_type(const char *tag, int taglen, const char *type,
		char **p)
{
	char *q = key_value_copy(tag, taglen, type);
	if (!q)
		return false;
	free(*p);
	*p = q;
	return true;
}

static int spx_get_comments(unsigned char *comments, int length,
		struct taginfo *tags)
{
	char *c = (char *)comments;
	uint32_t len, nb_fields;
	int i;
	char *end;

	if (length < 8)
		return -E_SPX_COMMENT;
	end = c + length;
	len = read_u32(c);
	c += 4;
	if (c + len > end)
		return -E_SPX_COMMENT;
	tags->comment = safe_strdup(c, len);

	c += len;
	if (c + 4 > end)
		return -E_SPX_COMMENT;
	nb_fields = read_u32(c);
	PARA_DEBUG_LOG("%u comment(s)\n", nb_fields);
	c += 4;
	for (i = 0; i < nb_fields; i++, c += len) {
		char *tag;

		if (c + 4 > end)
			return -E_SPX_COMMENT;
		len = read_u32(c);
		c += 4;
		if (c + len > end)
			return -E_SPX_COMMENT;
		if (copy_if_tag_type(c, len, "author", &tags->artist))
			continue;
		if (copy_if_tag_type(c, len, "artist", &tags->artist))
			continue;
		if (copy_if_tag_type(c, len, "title", &tags->title))
			continue;
		if (copy_if_tag_type(c, len, "album", &tags->album))
			continue;
		if (copy_if_tag_type(c, len, "year", &tags->year))
			continue;
		if (copy_if_tag_type(c, len, "comment", &tags->comment))
			continue;
		tag = safe_strdup(c, len);
		PARA_NOTICE_LOG("unrecognized comment: %s\n", tag);
		free(tag);
	}
	return 1;
}

static int spx_packet_callback(ogg_packet *packet, int packet_num,
		__a_unused int serial, struct afh_info *afhi,
		void *private_data)
{
	struct private_spx_data *psd = private_data;
	int ret;

	if (packet_num == 0) {
		ret = spx_process_header(packet->packet, packet->bytes,
			&psd->shi);
		if (ret < 0)
			return ret;
		afhi->channels = psd->shi.channels;
		afhi->frequency = psd->shi.sample_rate;
		afhi->bitrate = psd->shi.bitrate / 1000;
		afhi->techinfo = make_message("%s, v%d", psd->shi.mode->modeName,
			psd->shi.mode->bitstream_version);
		return 1;
	}
	if (packet_num == 1) {
		ret = spx_get_comments(packet->packet, packet->bytes,
			&afhi->tags);
		if (ret < 0)
			return ret;
		return 0; /* header complete */
	}
	/* never reached */
	return 0;
}

static int spx_get_file_info(char *map, size_t numbytes, __a_unused int fd,
		struct afh_info *afhi)
{
	struct private_spx_data psd;
	struct oac_callback_info spx_callback_info = {
		.packet_callback = spx_packet_callback,
		.private_data = &psd,
	};

	memset(&psd, 0, sizeof(psd));
	return oac_get_file_info(map, numbytes, afhi, &spx_callback_info);
}

static size_t spx_make_meta_packet(struct taginfo *tags, char **result)
{
	size_t sz;
	char *buf, *p;
	size_t comment_len = strlen(tags->comment),
		artist_len = strlen(tags->artist),
		title_len = strlen(tags->title),
		album_len = strlen(tags->album),
		year_len = strlen(tags->year);
	uint32_t comment_sz = comment_len,
		artist_sz = artist_len + strlen("artist="),
		title_sz = title_len + strlen("title="),
		album_sz = album_len + strlen("album="),
		year_sz = year_len + strlen("year=");
	uint32_t num_tags;

	sz = 4 /* comment length (always present) */
		+ comment_sz
		+ 4; /* number of tags */
	num_tags = 0;
	if (artist_len) {
		num_tags++;
		sz += 4 + artist_sz;
	}
	if (title_len) {
		num_tags++;
		sz += 4 + title_sz;
	}
	if (album_len) {
		num_tags++;
		sz += 4 + album_sz;
	}
	if (year_len) {
		num_tags++;
		sz += 4 + year_sz;
	}
	PARA_DEBUG_LOG("meta packet size: %zu bytes\n", sz);
	/* terminating zero byte for the last sprintf() */
	buf = p = alloc(sz + 1);
	write_u32(p, comment_sz);
	p += 4;
	strcpy(p, tags->comment);
	p += comment_sz;
	write_u32(p, num_tags);
	p += 4;
	if (artist_len) {
		write_u32(p, artist_sz);
		p += 4;
		sprintf(p, "artist=%s", tags->artist);
		p += artist_sz;
	}
	if (title_len) {
		write_u32(p, title_sz);
		p += 4;
		sprintf(p, "title=%s", tags->title);
		p += title_sz;
	}
	if (album_len) {
		write_u32(p, album_sz);
		p += 4;
		sprintf(p, "album=%s", tags->album);
		p += album_sz;
	}
	if (year_len) {
		write_u32(p, year_sz);
		p += 4;
		sprintf(p, "year=%s", tags->year);
		p += year_sz;
	}
	assert(p == buf + sz);
	*result = buf;
	return sz;
}

static int spx_rewrite_tags(const char *map, size_t mapsize,
		struct taginfo *tags, int output_fd,
		__a_unused const char *filename)
{
	char *meta_packet;
	size_t meta_sz;
	int ret;

	meta_sz = spx_make_meta_packet(tags, &meta_packet);
	ret = oac_rewrite_tags(map, mapsize, output_fd, meta_packet, meta_sz);
	free(meta_packet);
	return ret;
}

static const char * const speex_suffixes[] = {"spx", "speex", NULL};

/**
 * The ogg/speex audio format handler.
 *
 * This codec is considered obsolete because the opus codec surpasses its
 * performance in all areas. It is only compiled in if both the ogg and the
 * speex library are installed.
 */
const struct audio_format_handler spx_afh = {
	.get_file_info = spx_get_file_info,
	.suffixes = speex_suffixes,
	.rewrite_tags = spx_rewrite_tags,
};
